from libc.stdio cimport FILE

from .atom cimport Atom
from .tess_atom cimport TessAtom
from .template cimport Template
from .molecule cimport Molecule
from .candidate_set cimport CandidateSet


cdef extern from "TessTemplate.h" nogil:

    cdef struct _TessTemplate:
        int count
        TessAtom** atom
        double** distance
        char* symbol
        int dim

    ctypedef _TessTemplate TessTemplate

    int TessTemplate_count(const Template *T)
    int TessTemplate_match(const Template *T,int k,const Atom *A)
    int TessTemplate_range(const Template *T,int i,int j,double *a,double *b)
    const double *TessTemplate_position(const Template *T, int k)
    double TessTemplate_distWeight(const Template *T, int k)
    int TessTemplate_check(const Template *T, Atom **A, int k, int ignore_chain)
    CandidateSet* TessTemplate_candidates(const Template *T, const Molecule *M, int k)
    const char *TessTemplate_name(const Template *T)
    double TessTemplate_logE(const Template *T,double rmsd, int n)
    void TessTemplate_free(Template *T)

    Template* TessTemplate_create(FILE*, const char*)
    Template* TessTemplate_copy(Template*)




import math
import unittest
import sys
import pickle

from .._jess import Template, Molecule, Jess
from .utils import files
from . import data


class TestHit(unittest.TestCase):

    @unittest.skipUnless(files, "importlib.resources not available")
    @classmethod
    def setUpClass(cls):
        with files(data).joinpath("template_01.qry").open() as f:
            template = Template.load(f)
            jess = Jess([template])
        with files(data).joinpath("pdb1lnb.pdb").open() as f:
            molecule = Molecule.load(f)

        cls.hit = next(jess.query(molecule, 1, 2, 2))
        
    def test_pickle(self):
        hit = pickle.loads(pickle.dumps(self.hit))
        self.assertIsNot(self.hit, hit)
        self.assertEqual(self.hit.rmsd, hit.rmsd)
        self.assertEqual(self.hit.determinant, hit.determinant)
        self.assertEqual(self.hit.evalue, hit.evalue)
        self.assertEqual(self.hit.template, hit.template)
        self.assertListEqual(self.hit.atoms(transform=True), hit.atoms(transform=True))
        self.assertListEqual(self.hit.atoms(transform=False), hit.atoms(transform=False))
        self.assertEqual(self.hit.molecule(transform=False), hit.molecule(transform=False))


from .atom cimport Atom
from .molecule cimport Molecule
from .candidate_set cimport CandidateSet

cdef extern from "Template.h" nogil:

    ctypedef _Template Template
    cdef struct _Template:
        void (*free)(Template*) nogil
        int (*count)(const Template*) nogil
        int (*match)(const Template*, int, const Atom*) nogil
        int (*range)(const Template*, int, int, double*, double*) nogil
        int (*check)(const Template*, Atom**, int, int) nogil
        CandidateSet* (*candidates)(const Template*, const Molecule*, int);
        const double* (*position)(const Template*, int) nogil
        const char* (*name)(const Template*) nogil
        double (*logE)(const Template*, double, int) nogil
        double (*distWeight)(const Template*, int) nogil
        Template* (*copy)(const Template*) nogil



# coding: utf-8
# cython: language_level=3, linetrace=True, binding=True
"""Bindings to Jess, a 3D template matching software.

References:
    - Barker, J. A., & Thornton, J. M. (2003). *An algorithm for
      constraint-based structural template matching: application to
      3D templates with statistical analysis*. Bioinformatics (Oxford,
      England), 19(13), 1644–1649. :doi:`10.1093/bioinformatics/btg226`.
    - Wallace, A. C., Borkakoti, N., & Thornton, J. M. (1997).
      *TESS: a geometric hashing algorithm for deriving 3D coordinate
      templates for searching structural databases. Application to enzyme
      active sites*. Protein science : a publication of the Protein
      Society, 6(11), 2308–2323. :doi:`10.1002/pro.5560061104`.

"""

# --- C imports --------------------------------------------------------------

cimport cython
from cpython.unicode cimport PyUnicode_FromStringAndSize

from libc.math cimport isnan, exp, INFINITY, NAN
from libc.stdio cimport FILE, fclose, fdopen, printf
from libc.stdint cimport uintptr_t
from libc.stdlib cimport calloc, realloc, free, malloc
from libc.string cimport memcpy, memset, strncpy, strdup

cimport jess.atom
cimport jess.jess
cimport jess.molecule
cimport jess.super
cimport jess.tess_template
cimport jess.tess_atom
cimport jess.res_index
from jess.atom cimport Atom as _Atom
from jess.jess cimport Jess as _Jess
from jess.jess cimport JessQuery as _JessQuery
from jess.molecule cimport Molecule as _Molecule
from jess.super cimport Superposition as _Superposition
from jess.template cimport Template as _Template
from jess.tess_template cimport TessTemplate as _TessTemplate
from jess.tess_atom cimport TessAtom as _TessAtom

# --- Python imports ---------------------------------------------------------

import contextlib
import functools
import io
import itertools
import os
import warnings

__version__ = PROJECT_VERSION

# --- Utils ------------------------------------------------------------------

cdef inline void copy_token(char* dst, const char* src, size_t n) noexcept nogil:
    cdef size_t i
    for i in range(n):
        if src[i] == ord(' ') or src[i] == 0:
            dst[i] = ord('_')
        else:
            dst[i] = src[i]
    dst[n] = 0

@contextlib.contextmanager
def nullcontext(return_value=None):
    yield return_value

# --- Classes ----------------------------------------------------------------

cdef class Molecule:
    """A molecule structure, as a sequence of `Atom` objects.

    .. versionadded:: 0.2.2
       Support identifiers of arbitrary length.

    .. versionadded:: 0.4.0
       Equality, hashing and pickle protocol support.

    """
    cdef _Molecule* _mol
    cdef str        _id

    @classmethod
    def loads(cls, text, str id = None, bint ignore_endmdl = False):
        """Load a molecule from a PDB string.

        Arguments:
            file (`str`, `os.PathLike`, or file-like object): Either the path
                to a file, or a file-like object opened in **text mode**
                containing a PDB molecule.
            id (`str`, optional): The identifier of the molecule. If `None`
                given, the parser will attempt to extract it from the
                ``HEADER`` line.
            ignore_endmdl (`bool`): Pass `True` to make the parser read all
                the atoms from the PDB file. By default, the parser only
                reads the atoms of the first model, and stops at the first
                ``ENDMDL`` line.

        Returns:
            `~pyjess.Molecule`: The molecule parsed from the PDB file.

        See Also:
            `Molecule.load` to load a PDB molecule from a file-like
            object or from a path.

        """
        return cls.load(io.StringIO(text), id=id, ignore_endmdl=ignore_endmdl)

    @classmethod
    def load(cls, file, str id = None, bint ignore_endmdl = False):
        """Load a molecule from a PDB file.

        Arguments:
            file (`str`, `os.PathLike`, or file-like object): Either the path
                to a file, or a file-like object opened in **text mode**
                containing a PDB molecule.
            id (`str`, optional): The identifier of the molecule. If `None`
                given, the parser will attempt to extract it from the
                ``HEADER`` line.
            ignore_endmdl (`bool`): Pass `True` to make the parser read all
                the atoms from the PDB file. By default, the parser only
                reads the atoms of the first model, and stops at the first
                ``ENDMDL`` line.

        Returns:
            `~pyjess.Molecule`: The molecule parsed from the PDB file.

        """
        try:
            handle = open(file)
        except TypeError:
            handle = nullcontext(file)
        with handle as f:
            atoms = []
            for line in f:
                if line.startswith("HEADER"):
                    if id is None:
                        id = line[62:66].strip() or None
                elif line.startswith(("ATOM", "HETATM")):
                    atoms.append(Atom.loads(line))
                elif line.startswith("ENDMDL"):
                    if not ignore_endmdl:
                        break
        return cls(atoms, id=id)

    def __cinit__(self):
        self._mol = NULL

    def __dealloc__(self):
        jess.molecule.Molecule_free(self._mol)

    def __init__(self, object atoms = (), str id = None):
        """__init__(self, atoms=(), id=None)\n--\n

        Create a new molecule.

        Arguments:
            atoms (sequence of `~pyjess.Atom`): The atoms of the molecule.
            id (`str`, optional): The identifier of the molecule.

        Raises:
            `MemoryError`: When the system allocator fails to allocate
                enough memory for the molecule storage.

        """
        cdef Atom atom
        cdef int i
        cdef int count = len(atoms)

        self._mol = <_Molecule*> malloc(sizeof(_Molecule) + count * sizeof(_Atom*))
        if self._mol is NULL:
            raise MemoryError("Failed to allocate molecule")

        self._mol.index = NULL
        self._mol.count = count
        for i in range(count):
            self._mol.atom[i] = NULL
        memset(self._mol.id, b' ', 5)
        self._id = id

        for i, atom in enumerate(atoms):
            self._mol.atom[i] = <_Atom*> malloc(sizeof(_Atom))
            if self._mol.atom[i] is NULL:
                raise MemoryError("Failed to allocate atom")
            memcpy(self._mol.atom[i], atom._atom, sizeof(_Atom))

        self._mol.index = jess.res_index.ResIndex_create(self._mol.atom, count)
        if self._mol.index is NULL:
            raise MemoryError("Failed to allocate residue index")

    def __len__(self):
        assert self._mol is not NULL
        return self._mol.count

    def __getitem__(self, object index):
        assert self._mol is not NULL

        cdef Atom    atom
        cdef ssize_t index_
        cdef ssize_t length = self._mol.count

        if isinstance(index, slice):
            indices = range(*index.indices(length))
            return type(self)(atoms=[self[i] for i in indices], id=self.id)
        else:
            index_ = index
            if index_ < 0:
                index_ += length
            if index_ < 0 or index_ >= length:
                raise IndexError(index)
            atom = Atom.__new__(Atom)
            atom.owner = self
            atom.owned = True
            atom._atom = <_Atom*> jess.molecule.Molecule_atom(self._mol, index_)
            return atom

    def __copy__(self):
        return self.copy()

    def __eq__(self, object other):
        cdef Molecule other_
        if not isinstance(other, Molecule):
            return NotImplemented
        other_ = other
        if self._id != other_._id:
            return False
        if self._mol.count != other_._mol.count:
            return False
        return all(x == y for x,y in zip(self, other_))

    def __hash__(self):
        return hash((self._id, *(hash(x) for x in self)))

    def __reduce__(self):
        return type(self), (list(self), self.id)

    def __sizeof__(self):
        assert self._mol is not NULL
        return (
            sizeof(self)
            + sizeof(_Molecule)
            + self._mol.count*(sizeof(_Atom*) + sizeof(_Atom))
        )

    @property
    def id(self):
        return self._id

    cpdef Molecule conserved(self, double cutoff = 0.0):
        assert self._mol is not NULL
        cdef Atom atom
        return type(self)(
            id=self.id,
            atoms=[
                atom
                for atom in self
                if cutoff <= 0.0
                or atom._atom.tempFactor >= cutoff
            ]
        )

    cpdef Molecule copy(self):
        """Create a copy of this molecule and its atoms.

        Returns:
            `~pyjess.Molecule`: A newly allocated molecule with the same
            identifier and atoms.

        .. versionadded:: 0.4.0

        """
        cdef Molecule copy = Molecule.__new__(Molecule)
        cdef size_t   size = sizeof(_Molecule) + self._mol.count * sizeof(_Atom*)

        with nogil:
            # allocate molecule storage
            copy._mol = <_Molecule*> malloc(size)
            if copy._mol is NULL:
                raise MemoryError("Failed to allocate molecule")
            # copy molecule attributes
            copy._mol.index = NULL
            copy._mol.count = self._mol.count
            memset(copy._mol.id, b' ', 5)
            # copy molecule atoms
            for i in range(self._mol.count):
                copy._mol.atom[i] = <_Atom*> malloc(sizeof(_Atom))
                if copy._mol.atom[i] is NULL:
                    raise MemoryError("Failed to allocate atom")
                memcpy(copy._mol.atom[i], self._mol.atom[i], sizeof(_Atom))
            # regenerate index
            copy._mol.index = jess.res_index.ResIndex_create(copy._mol.atom, copy._mol.count)
            if copy._mol.index is NULL:
                raise MemoryError("Failed to allocate residue index")

        copy._id = self._id
        return copy


cdef class Atom:
    """A single atom in a molecule.

    .. versionadded:: 0.4.0
       Equality, hashing and pickle protocol support.

    """
    cdef object owner
    cdef bint   owned
    cdef _Atom* _atom

    @classmethod
    def load(cls, file):
        """Load an atom from the given file.

        Arguments:
            file (file-like object): A file-like object opened in text
                mode to read the atom from.

        """
        return cls.loads(file.read())

    @classmethod
    def loads(cls, text):
        """Load an atom from the given string.

        Arguments:
            text (`str`, `bytes` or `bytearray`): The atom line to read the
                atom metadata from.

        """
        cdef const unsigned char* s
        cdef bytearray            b
        cdef Atom                 atom

        if isinstance(text, str):
            b = bytearray(text, 'utf-8')
        else:
            b = bytearray(text)
        if not b.endswith(b'\n'):
            b.append(b'\n')
        b.append(b'\0')
        s = b

        atom = cls.__new__(cls)
        with nogil:
            atom._atom = <_Atom*> malloc(sizeof(_Atom))
            if atom._atom == NULL:
                raise MemoryError("Failed to allocate atom")
            if not jess.atom.Atom_parse(atom._atom, <const char*> s):
                raise ValueError(f"Failed to parse atom: {text!r}")

        return atom

    def __cinit__(self):
        self._atom = NULL
        self.owner = None
        self.owned = False

    def __dealloc__(self):
        if not self.owned:
            free(self._atom)

    def __init__(
        self,
        *,
        int serial,
        str name,
        str altloc,
        str residue_name,
        str chain_id,
        int residue_number,
        str insertion_code,
        double x,
        double y,
        double z,
        double occupancy = 0.0,
        double temperature_factor = 0.0,
        str segment = '',
        str element = '',
        int charge = 0,
    ):
        """__init__(self, *, serial, name, altloc, residue_name, chain_id, residue_number, insertion_code, x, y, z, occupancy=0.0, temperature_factor=0.0, segment='', element='', charge=0)\n--\n

        Create a new atom.

        Raises:
            `MemoryError`: When the system allocator fails to allocate
                enough memory for the atom storage.
            `ValueError`: When either of the ``name``, ``residue_name``,
                ``segment``, ``element`` or ``chain_id`` strings is too
                long.

        """
        if len(name) > 4:
            raise ValueError(f"Invalid atom name: {name!r}")
        if len(residue_name) > 3:
            raise ValueError(f"Invalid residue name: {residue_name!r}")
        if len(segment) > 3:
            raise ValueError(f"Invalid segment: {segment!r}")
        if len(element) > 2:
            raise ValueError(f"Invalid element: {element!r}")
        if len(chain_id) > 2:
            raise ValueError(f"Invalid chain ID: {chain_id!r}")

        self._atom = <_Atom*> malloc(sizeof(_Atom))
        if self._atom is NULL:
            raise MemoryError("Failed to allocate atom")

        self._atom.serial = serial
        self._atom.altLoc = ord(altloc)
        self._atom.chainID1 = ord(chain_id[0]) if len(chain_id) > 0 else 0
        self._atom.chainID2 = ord(chain_id[1]) if len(chain_id) > 1 else ord(' ')
        self._atom.resSeq = residue_number
        self._atom.iCode = ord(insertion_code)
        self._atom.x[0] = x
        self._atom.x[1] = y
        self._atom.x[2] = z
        self._atom.occupancy = occupancy
        self._atom.tempFactor = temperature_factor
        self._atom.charge = charge
        copy_token(self._atom.resName, residue_name.encode('ascii').ljust(3, b'\0'), 3)
        copy_token(self._atom.segID, segment.encode('ascii').ljust(3, b'\0'), 3)
        copy_token(self._atom.element, element.encode('ascii').ljust(2, b'\0'), 2)

        _name = bytearray(name, 'ascii')
        if len(_name) < 4:
            _name.insert(0, ord('_'))
        copy_token(self._atom.name, _name.ljust(4, b'\0'), 4)

    def __copy__(self):
        return self.copy()

    cdef dict _state(self):
        return {
            "serial": self.serial,
            "name": self.name,
            "altloc": self.altloc,
            "residue_name": self.residue_name,
            "chain_id": self.chain_id,
            "residue_number": self.residue_number,
            "insertion_code": self.insertion_code,
            "x": self.x,
            "y": self.y,
            "z": self.z,
            "temperature_factor": self.temperature_factor,
            "occupancy": self.occupancy,
            "segment": self.segment,
            "element": self.element,
            "charge": self.charge,
        }

    def __reduce__(self):
        cdef dict state = self._state()
        return functools.partial(type(self), **state), ()

    def __repr__(self):
        cdef str  ty   = type(self).__name__
        cdef list args = []
        for k,v in self._state().items():
            if v is not None:
                args.append(f"{k}={v!r}")
        return f"{ty}({', '.join(args)})"

    def __sizeof__(self):
        cdef size_t size = sizeof(self)
        if not self.owned:
            size += sizeof(_Atom)
        return size

    def __eq__(self, object other):
        cdef Atom other_
        if not isinstance(other, Atom):
            return NotImplemented
        other_ = other
        # FIXME: it should be possible to do a memcmp here.
        return self._state() == other_._state()

    def __hash__(self):
        return hash(tuple(self._state().values()))

    @property
    def serial(self):
        """`int`: The atom serial number.
        """
        assert self._atom is not NULL
        return self._atom.serial

    @property
    def altloc(self):
        """`str`: The alternate location indicator for the atom.
        """
        assert self._atom is not NULL
        return chr(self._atom.altLoc)

    @property
    def name(self):
        """`str`: The atom name.
        """
        assert self._atom is not NULL
        return self._atom.name[:4].decode('ascii').strip("_")

    @property
    def residue_name(self):
        """`str`: The residue name.
        """
        assert self._atom is not NULL
        return self._atom.resName[:3].decode('ascii').strip("_")

    @property
    def residue_number(self):
        """`int`: The residue sequence number.
        """
        assert self._atom is not NULL
        return self._atom.resSeq

    @property
    def segment(self):
        """`str`: The segment identifier.
        """
        assert self._atom is not NULL
        return self._atom.segID[:3].decode('ascii').strip('_')

    @property
    def element(self):
        """`str`: The element symbol.
        """
        assert self._atom is not NULL
        return self._atom.element[:2].decode('ascii').strip('_')

    @property
    def insertion_code(self):
        """`str`: The code for insertion of residues.
        """
        assert self._atom is not NULL
        return chr(self._atom.iCode)

    @property
    def chain_id(self):
        """`str`: The identifier of the chain the atom belongs to.
        """
        assert self._atom is not NULL
        return "{}{}".format(chr(self._atom.chainID1), chr(self._atom.chainID2)).strip()

    @property
    def occupancy(self):
        """`float`: The atom occupancy.
        """
        assert self._atom is not NULL
        return self._atom.occupancy

    @property
    def temperature_factor(self):
        """`float`: The atom temperature factor.
        """
        assert self._atom is not NULL
        return self._atom.tempFactor

    @property
    def charge(self):
        """`int`: The atom charge.
        """
        assert self._atom is not NULL
        return self._atom.charge

    @property
    def x(self):
        assert self._atom is not NULL
        return self._atom.x[0]

    @property
    def y(self):
        assert self._atom is not NULL
        return self._atom.x[1]

    @property
    def z(self):
        assert self._atom is not NULL
        return self._atom.x[2]

    cpdef Atom copy(self):
        """Create a copy of this atom.

        Returns:
            `~pyjess.Atom`: A newly allocated atom with identical attributes.

        .. versionadded:: 0.4.0

        """
        cdef Atom copy = Atom.__new__(Atom)
        copy._atom = <_Atom*> malloc(sizeof(_Atom))
        if copy._atom is NULL:
            raise MemoryError("Failed to allocate atom")
        memcpy(copy._atom, self._atom, sizeof(_Atom))
        return copy


cdef class TemplateAtom:
    """A single template atom.

    .. versionadded:: 0.4.0
       Equality, hashing and pickle protocol support.

    """
    cdef object     owner
    cdef bint       owned
    cdef _TessAtom* _atom

    @classmethod
    def load(cls, file):
        """Load a template atom from the given file.

        Arguments:
            file (str, os.PathLike or file-like object): A file-like object
                opened in text or binary mode to read the template atom from.

        """
        try:
            handle = open(file)
        except TypeError:
            handle = nullcontext(file)
        with handle as f:
            return cls.loads(f.read())

    @classmethod
    def loads(cls, text):
        """Load a template atom from the given string.

        Arguments:
            text (`str`, `bytes` or `bytearray`): The atom line to read the
                atom metadata from.

        """
        cdef bytearray    b
        cdef TemplateAtom atom

        if isinstance(text, str):
            b = bytearray(text, 'utf-8')
        else:
            b = bytearray(text)
        if not b.endswith(b'\n'):
            b.append(b'\n')
        b.append(b'\0')

        atom = TemplateAtom.__new__(TemplateAtom)
        atom._atom = jess.tess_atom.TessAtom_create(<const char*> b)
        if atom._atom == NULL:
            raise ValueError(f"Failed to parse template atom: {text!r}")

        # validate match mode *now* to avoid Jess exiting when it does so later
        if atom.match_mode not in range(-1, 9) and atom.match_mode not in range(100, 108):
            raise ValueError(f"Invalid match mode: {atom.match_mode!r}")

        return atom

    def __cinit__(self):
        self.owner = None
        self.owned = False
        self._atom = NULL

    def __dealloc__(self):
        if not self.owned:
            jess.tess_atom.TessAtom_free(self._atom)

    def __init__(
        self,
        *,
        str chain_id,
        int residue_number,
        double x,
        double y,
        double z,
        object residue_names,
        object atom_names,
        double distance_weight = 0.0,
        int match_mode = 0,
    ):
        """__init__(self, *, chain_id, residue_number, x, y, z, residue_names, atom_names, distance_weight=0.0, match_mode=0)\n--\n

        Create a new template atom.

        Raises:
            `MemoryError`: When the system allocator fails to allocate
                enough memory for the template atom storage.

        """
        cdef size_t m
        cdef char*  p
        cdef size_t ac
        cdef size_t rc
        cdef size_t alloc_size

        # validate match mode to avoid a potential hard exit later
        if match_mode not in range(-1, 9) and match_mode not in range(100, 108):
            raise ValueError(f"Invalid match mode: {match_mode!r}")
        if len(chain_id) > 2:
            raise ValueError(f"Invalid chain ID: {chain_id!r}")

        # compute total allocation
        ac = len(atom_names)
        rc = len(residue_names)
        alloc_size = sizeof(_TessAtom) + sizeof(char*) * (ac + rc) + sizeof(char) * (5*ac + 4*rc)

        # allocate base memory
        self._atom = <_TessAtom*> malloc(alloc_size)
        if self._atom is NULL:
            raise MemoryError("Failed to allocate template atom")

        # copy base data
        self._atom.code = match_mode
        self._atom.resSeq = residue_number
        self._atom.pos[0] = x
        self._atom.pos[1] = y
        self._atom.pos[2] = z
        self._atom.chainID1, self._atom.chainID2 = map(ord, chain_id.ljust(2))
        self._atom.nameCount = ac
        self._atom.resNameCount = rc
        self._atom.distWeight = distance_weight

        # setup string pointers
        p = <char*> &self._atom[1]
        self._atom.name = <char**> p
        p += sizeof(char*)*ac
        for m in range(ac):
            self._atom.name[m] = <char*> p
            p += 5
        self._atom.resName = <char**> p
        p += sizeof(char*)*rc
        for m in range(rc):
            self._atom.resName[m] = <char*> p
            p += 4

        # copy atom names
        for m, name in enumerate(atom_names):
            if isinstance(name, str):
                _name = bytearray(name, 'ascii')
            else:
                _name = bytearray(name)
            if len(_name) > 4:
                raise ValueError(f"Invalid atom name: {name!r}")
            elif len(_name) < 3:
                _name.insert(0, ord('_'))
            copy_token(self._atom.name[m], _name.ljust(4, b'\0'), 4)

        # copy residue names
        for m, name in enumerate(residue_names):
            _name = name.encode('ascii') if isinstance(name, str) else name
            if len(_name) > 3:
                raise ValueError(f"Invalid residue name: {name!r}")
            copy_token(self._atom.resName[m], _name.ljust(3, b'\0'), 3)

    cdef dict _state(self):
        return {
            "chain_id": self.chain_id,
            "residue_number": self.residue_number,
            "x": self.x,
            "y": self.y,
            "z": self.z,
            "residue_names": self.residue_names,
            "atom_names": self.atom_names,
            "distance_weight": self.distance_weight,
            "match_mode": self.match_mode,
        }

    def __repr__(self):
        cdef str  ty   = type(self).__name__
        cdef list args = []
        for k, v in self._state().items():
            args.append(f"{k}={v!r}")
        return f"{ty}({', '.join(args)})"

    def __copy__(self):
        return self.copy()

    def __eq__(self, object other):
        cdef TemplateAtom other_
        if not isinstance(other, TemplateAtom):
            return NotImplemented
        other_ = other
        return self._state() == other_._state()

    def __hash__(self):
        return hash(tuple(self._state().values()))

    def __reduce__(self):
        return functools.partial(type(self), **self._state()), ()

    def __sizeof__(self):
        assert self._atom is not NULL

        cdef size_t ac   = self._atom.nameCount
        cdef size_t rc   = self._atom.resNameCount
        cdef size_t size = sizeof(self)

        if not self.owned:
            size += (
                sizeof(_TessAtom)
                + sizeof(char*) * (ac + rc)
                + sizeof(char) * (5*ac + 4*rc)
            )
        return size

    @property
    def match_mode(self):
        """`int`: The match mode for this particular atom.
        """
        assert self._atom is not NULL
        return self._atom.code

    @property
    def residue_number(self):
        """`int`: The residue sequence number.
        """
        assert self._atom is not NULL
        return self._atom.resSeq

    @property
    def chain_id(self):
        """`str`: The identifier of the chain the atom belongs to.
        """
        assert self._atom is not NULL
        cdef char c1 = jess.tess_atom.TessAtom_chainID1(self._atom)
        cdef char c2 = jess.tess_atom.TessAtom_chainID2(self._atom)
        return "{}{}".format(chr(c1), chr(c2)).strip()

    @property
    def x(self):
        """`float`: The x coordinate of the atom.
        """
        assert self._atom is not NULL
        return self._atom.pos[0]

    @property
    def y(self):
        """`float`: The y coordinate of the atom.
        """
        assert self._atom is not NULL
        return self._atom.pos[1]

    @property
    def z(self):
        """`float`: The z coordinate of the atom.
        """
        assert self._atom is not NULL
        return self._atom.pos[2]

    @property
    def atom_names(self):
        """`tuple` of `str`: The different atom names for this atom.

        .. versionchanged:: 0.4.1
           Property now returns a `tuple` rather than a `list`.

        """
        assert self._atom is not NULL

        cdef int  i
        cdef list l = []

        for i in range(self._atom.nameCount):
            l.append(self._atom.name[i].replace(b'_', b'').decode())
        return tuple(l)

    @property
    def residue_names(self):
        """`tuple` of `str`: The different residue names for this atom.

        .. versionchanged:: 0.4.1
           Property now returns a `tuple` rather than a `list`.

        """
        assert self._atom is not NULL

        cdef int  i
        cdef list l = []

        for i in range(self._atom.resNameCount):
            l.append(self._atom.resName[i].replace(b'_', b'').decode())
        return tuple(l)

    @property
    def distance_weight(self):
        """`float`: The distance weight for this atom.
        """
        assert self._atom is not NULL
        return self._atom.distWeight

    cpdef TemplateAtom copy(self):
        """Create a copy of this template atom.

        Returns:
            `~pyjess.TemplateAtom`: A new template atom object with
            identical attributes.

        .. versionadded:: 0.4.0

        """
        return type(self)(**self._state())


cdef class Template:
    """A template, as a sequence of `TemplateAtom` objects.

    .. versionadded:: 0.4.0
       Equality, hashing and pickle protocol support.

    """
    cdef object         owner
    cdef bint           owned
    cdef _Template*     _tpl
    cdef _TessTemplate* _tess

    @classmethod
    def loads(cls, text, str id = None):
        """Load a template from a string.

        Arguments:
            file (`str`, `os.PathLike`, or file-like object): Either the path
                to a file, or a file-like object opened in **text mode**
                containing the template.
            id (`str`, optional): The identifier of the template. By default,
                the parser will take the one from the ``PDB_ID`` remark if
                found in the header.

        Returns:
            `~pyjess.Template`: The template parsed from the given string.

        See Also:
            `Template.load` to load a template from a file-like object or
            from a path.

        """
        return cls.load(io.StringIO(text), id=id)

    @classmethod
    def load(cls, file, str id = None):
        """Load a template from the given file.

        Arguments:
            file (`str`, `os.PathLike` or file-like object): Either the
                path to a file, or a file-like object opened in **text mode**
                to read the template from.
            id (`str`, optional): The identifier of the template. By default,
                the parser will take the one from the ``PDB_ID`` remark if
                found in the header.

        Returns:
            `~pyjess.Template`: The template parsed from the given file.

        """
        try:
            handle = open(file)
        except TypeError:
            handle = nullcontext(file)
        with handle as f:
            atoms = []
            for line in f:
                if line.startswith("ATOM"):
                    atoms.append(TemplateAtom.loads(line))
                elif id is None and line.startswith("REMARK PDB_ID"):
                    id = line.split(" ", maxsplit=2)[2].strip()
        return cls(atoms, id=id)

    def __cinit__(self):
        self._tpl  = NULL
        self._tess = NULL
        self.owner = None
        self.owned = False

    def __dealloc__(self):
        if not self.owned:
            jess.tess_template.TessTemplate_free(self._tpl)

    def __init__(self, object atoms = (), str id = None):
        """__init__(self, atoms=(), id=None)\n--\n

        Create a new template.

        Arguments:
            atoms (sequence of `~pyjess.TemplateAtom`): The atoms of the
                templates.
            id (`str`, optional): The identifier of the template.

        Raises:
            `MemoryError`: When the system allocator fails to allocate
                enough memory for the template storage.

        """
        cdef int          i
        cdef int          j
        cdef double       dist
        cdef TemplateAtom atom
        cdef size_t       alloc_size
        cdef int          count      = len(atoms)

        alloc_size = (
            sizeof(_Template) + sizeof(_TessTemplate)
            + count * sizeof(_TessAtom*)
            + count * sizeof(double*)
            + count * count * sizeof(double)
        )

        self._tpl = <_Template*> calloc(1, alloc_size)
        if self._tpl is NULL:
            raise MemoryError("Failed to allocate template")

        # setup memory for atoms
        self._tess = <_TessTemplate*> &self._tpl[1]
        self._tess.atom = <_TessAtom**> &self._tess[1]
        for i in range(count):
            self._tess.atom[i] = NULL

        # setup memory and pointers for distances
        self._tess.distance = <double**> &self._tess.atom[count]
        if count > 0:
            self._tess.distance[0] = <double*> &self._tess.distance[count]
        for i in range(1, count):
            self._tess.distance[i] = <double*> &self._tess.distance[i-1][count]

        # setup template function pointers
        self._tpl.free = jess.tess_template.TessTemplate_free
        self._tpl.match = jess.tess_template.TessTemplate_match
        self._tpl.position = jess.tess_template.TessTemplate_position
        self._tpl.count = jess.tess_template.TessTemplate_count
        self._tpl.range = jess.tess_template.TessTemplate_range
        self._tpl.check = jess.tess_template.TessTemplate_check
        self._tpl.candidates = jess.tess_template.TessTemplate_candidates
        self._tpl.name = jess.tess_template.TessTemplate_name
        self._tpl.logE = jess.tess_template.TessTemplate_logE
        self._tpl.distWeight = jess.tess_template.TessTemplate_distWeight
        self._tpl.copy = jess.tess_template.TessTemplate_copy

        # copy name and atom count
        self._tess.count = count
        self._tess.symbol = NULL if id is None else strdup(id.encode())

        # copy atom data
        for i, atom in enumerate(atoms):
            assert i < count
            self._tess.atom[i] = jess.tess_atom.TessAtom_copy(atom._atom)
            if self._tess.atom[i] is NULL:
                raise MemoryError("Failed to allocate template atom")

        # compute distances
        for i in range(count):
            self._tess.distance[i][i] = 0.0
            for j in range(i+1, count):
                dist = jess.tess_atom.TessAtom_distance(self._tess.atom[i], self._tess.atom[j])
                self._tess.distance[i][j] = dist
                self._tess.distance[j][i] = dist

        # compute dimension
        residues = { 
            (
                self._tess.atom[i].resSeq , 
                self._tess.atom[i].chainID1,
                self._tess.atom[i].chainID2,
            )
            for i in range(count) 
        }
        self._tess.dim = len(residues)

    def __copy__(self):
        return self.copy()

    def __len__(self):
        assert self._tpl is not NULL
        return self._tess.count

    def __getitem__(self, object index):
        assert self._tess is not NULL

        cdef TemplateAtom atom
        cdef ssize_t      length = self._tess.count
        cdef ssize_t      index_

        if isinstance(index, slice):
            indices = range(*index.indices(length))
            return type(self)(atoms=[self[i] for i in indices], id=self.id)
        else:
            index_ = index
            if index_ < 0:
                index_ += length
            if index_ < 0 or index_ >= length:
                raise IndexError(index)
            atom = TemplateAtom.__new__(TemplateAtom)
            atom.owner = self
            atom.owned = True
            atom._atom = self._tess.atom[index_]
            return atom

    def __eq__(self, object other):
        cdef Template other_
        if not isinstance(other, Template):
            return NotImplemented
        other_ = other
        if self.id != other_.id:
            return False
        if self.dimension != other_.dimension:
            return False
        if len(self) != len(other_):
            return False
        return all(x == y for x,y in zip(self, other_))

    def __hash__(self):
        return hash((
            self.id,
            *(hash(x) for x in self)
        ))

    def __reduce__(self):
        return type(self), (list(self), self.id)

    def __sizeof__(self):
        assert self._tess is not NULL

        cdef size_t     i
        cdef size_t     ac
        cdef size_t     rc
        cdef _TessAtom* atom
        cdef size_t     size = sizeof(self)

        size = (
            sizeof(_Template)
            + sizeof(_TessTemplate)
            + self._tess.count * sizeof(_TessAtom*)
            + self._tess.count * sizeof(double*)
            + self._tess.count * self._tess.count * sizeof(double)
        )
        for i in range(self._tess.count):
            atom = self._tess.atom[i]
            ac = atom.nameCount
            rc = atom.resNameCount
            size += (
                sizeof(_TessAtom)
                + sizeof(char*) * (ac + rc)
                + sizeof(char) * (5*ac + 4*rc)
            )
        return size

    @property
    def id(self):
        assert self._tpl is not NULL

        cdef const char* name = self._tpl.name(self._tpl)
        if name is NULL:
            return None
        return name.decode()

    @property
    def dimension(self):
        """`int`: The dimension of the template (i.e. number of residues).
        """
        assert self._tess is not NULL
        return self._tess.dim

    cpdef Template copy(self):
        return Template(
            self,
            self.id
        )


cdef class Query:
    """A query over templates with a given molecule.

    Jess iterates over the templates and attempt matches the query
    molecule, so the hits can actually be generated iteratively. This
    class allows accessing the hits as a Python iterator.

    Attributes:
        jess (`~pyjess.Jess`): The templates this object is currently
            scanning.
        molecule (`~pyjess.Molecule`): The query molecule to align to
            the templates.
        rmsd_threshold (`float`): The RMSD threshold for reporting
            results.
        ignore_chain (`bool`): Whether to check or ignore the chain of
            the atoms to match.
        best_match (`bool`): Whether the query will return only the
            best match to each template.

    """
    cdef _JessQuery* _jq
    cdef bint        _partial
    cdef int         _candidates
    cdef uintptr_t   _prev_tpl
    cdef int         _max_candidates

    cdef readonly Jess     jess
    cdef readonly Molecule molecule
    cdef readonly bint     ignore_chain
    cdef readonly bint     best_match
    cdef readonly double   rmsd_threshold

    def __cinit__(self):
        self._jq = NULL
        self._candidates = 0
        self._partial = False
        self._prev_tpl = 0

    def __dealloc__(self):
        jess.jess.JessQuery_free(self._jq)

    def __iter__(self):
        return self

    @property
    def max_candidates(self):
        """`int`: The maximum number of candidate hits to report *by template*.
        """
        return None if self._max_candidates == -1 else self._max_candidates

    @max_candidates.setter
    def max_candidates(self, max_candidates):
        if max_candidates is None:
            self._max_candidates = -1
        elif max_candidates >= 0:
            self._max_candidates = max_candidates
        else:
            raise ValueError(f"invalid value for `max_candidates` argument: {max_candidates!r}")

    cdef bint _advance(self) noexcept nogil:
        if self._partial:
            self._partial = False
            return True
        return jess.jess.JessQuery_next(self._jq, self.ignore_chain)

    cdef bint _rewind(self) noexcept nogil:
        self._partial = True

    cdef int _copy_atoms(self, const _Template* tpl, Hit hit) except -1 nogil:
        cdef _Atom** atoms = jess.jess.JessQuery_atoms(self._jq)
        cdef int     count = tpl.count(tpl)

        hit._atoms = <_Atom*> realloc(hit._atoms, count * sizeof(_Atom))
        if hit._atoms is NULL:
            raise MemoryError("Failed to allocate hit atoms")
        for i in range(count):
            memcpy(&hit._atoms[i], atoms[i], sizeof(_Atom))
        return count

    cdef int _copy_superposition(self, _Superposition* sup, Hit hit) noexcept nogil:
        cdef const double* M = jess.super.Superposition_rotation(sup)
        cdef const double* c = jess.super.Superposition_centroid(sup, 0)
        cdef const double* v = jess.super.Superposition_centroid(sup, 1)
        memcpy(hit._rotation, M, 9*sizeof(double))
        memcpy(hit._centre[0], c, 3*sizeof(double))
        memcpy(hit._centre[1], v, 3*sizeof(double))
        return 0

    def __next__(self):
        assert self._jq is not NULL

        cdef double          rmsd
        cdef const double*   rot
        cdef _Template*      tpl     = NULL
        cdef _Template*      hit_tpl = NULL
        cdef _Superposition* sup     = NULL
        cdef Hit             hit     = Hit.__new__(Hit)

        # prepare the hit to be returned
        hit.rmsd = INFINITY
        hit._atoms = NULL
        hit._molecule = self.molecule
        hit_tpl = NULL
        hit_found = False

        # search the next hit without the GIL to allow parallel queries.
        with nogil:
            while self._advance():
                # load current iteration template, and check that the hit
                # was obtained with the current template and not with the
                # previous one
                self._prev_tpl = <uintptr_t> tpl
                tpl = jess.jess.JessQuery_template(self._jq)
                if hit_found and hit_tpl != tpl:
                    self._rewind()
                    break

                # check if we already made it to the next template,
                # or if we need to short-circuit the iteration and
                # force the query to move to the next template as
                # we found too many candidates already.
                if <uintptr_t> tpl != self._prev_tpl:
                    self._candidates = 0 
                else:
                    self._candidates += 1
                if self._candidates == self._max_candidates:
                    jess.jess.JessQuery_nextTemplate(self._jq)
                    continue

                # load superposition and compute RMSD for the current iteration
                sup = jess.jess.JessQuery_superposition(self._jq)
                rmsd = jess.super.Superposition_rmsd(sup)

                # NB(@althonos): we don't need to compute the E-value to get the
                #                best match by molecule/template pair since the
                #                logE-value for a fixed pair varies by the RMSD
                #                term only (see `TessTemplate_logE`)

                # check that the candidate passes threshold, and return it
                # if not in best match, otherwise record it until the next
                # template is reached (or the iterator finished)
                if rmsd <= self.rmsd_threshold and rmsd < hit.rmsd:
                    # check if the rotation matrix contains NaN values
                    rot = jess.super.Superposition_rotation(sup)
                    nan = False
                    for i in range(9):
                        nan |= isnan(rot[i])

                    if nan:
                        with gil:
                            warnings.warn(
                                "Jess returned a superposition matrix with NaN values",
                                UserWarning,
                                stacklevel=2,
                            )
                    else:
                        self._copy_atoms(tpl, hit)
                        self._copy_superposition(sup, hit)
                        hit.rmsd = rmsd
                        hit_tpl = tpl
                        hit_found = True

                # free superposition items (as relevant data was copied in
                # the Hit if needed) and return hits immediately if we are
                # not in best match mode
                jess.super.Superposition_free(sup)
                if hit_found and not self.best_match:
                    break

        if not hit_found:
            raise StopIteration

        # get the template object for the hit
        hit.template = self.jess._templates[self.jess._indices[<size_t> hit_tpl]]
        return hit


cdef class Hit:
    """A hit identified between a query molecule and a target template.

    Attributes:
        rmsd (`float`): The RMSD between the aligned structures.
        template (`~pyjess.Template`): The template that matched the
            query molecule.
        molecule (`~pyjess.Molecule`): The query molecule.

    """
    cdef double[9]       _rotation
    cdef double[2][3]    _centre
    cdef _Atom*          _atoms

    cdef readonly double   rmsd
    cdef readonly Template template
    cdef          Molecule _molecule

    def __dealloc__(self):
        free(self._atoms)

    def __getstate__(self):
        return {
            "rotation": list(self._rotation),
            "centre": list(self._centre),
            "atoms": self.atoms(transform=False),
            "rmsd": self.rmsd,
            "template": self.template,
            "molecule": self.molecule(transform=False),
        }

    def __setstate__(self, state):
        cdef size_t i
        cdef size_t count
        cdef Atom   atom

        self.rmsd = state["rmsd"]
        self.template = state["template"]
        self._molecule = state["molecule"]
        self._rotation = state["rotation"]
        self._centre = state["centre"]

        # check number of atoms is consistent
        count = len(self.template)
        if len(state["atoms"]) != count:
            raise ValueError(f"unexpected number of atoms: {len(state['atoms'])!r} (expected {count!r})")
        # allocate or reallocate memory for atoms
        self._atoms = <_Atom*> realloc(self._atoms, count * sizeof(_Atom))
        if self._atoms is NULL:
            raise MemoryError("Failed to allocate hit atoms")
        # copy atom data
        for i, atom in enumerate(state["atoms"]):
            memcpy(&self._atoms[i], atom._atom, sizeof(_Atom))

    @property
    def determinant(self):
        """`float`: The determinant of the rotation matrix.
        """
        cdef const double* p   = self._rotation
        cdef double        det = 0.0

        with nogil:
            det += p[0] * (p[4] * p[8] - p[5] * p[7])
            det -= p[1] * (p[3] * p[8] - p[5] * p[6])
            det += p[2] * (p[3] * p[7] - p[4] * p[6])
        return det

    @property
    def log_evalue(self):
        """`float`: The logarithm of the E-value estimated for the hit.
        """
        assert self.template._tpl is not NULL

        cdef int    n
        cdef double e

        with nogil:
            n = jess.molecule.Molecule_count(self._molecule._mol)
            e = self.template._tpl.logE(self.template._tpl, self.rmsd, n)
        return e

    @property
    def evalue(self):
        """`float`: The E-value estimated for the hit.
        """
        cdef int    n
        cdef double e

        with nogil:
            n = jess.molecule.Molecule_count(self._molecule._mol)
            e = exp(self.template._tpl.logE(self.template._tpl, self.rmsd, n))
        return e

    cpdef list atoms(self, bint transform=True):
        """Get the list of query atoms matching the template.

        Arguments:
            transform (`bool`): Whether or not to transform coordinates
                of hits into template frame.

        Returns:
            `list` of `~pyjess.Atom`: The list of matching atoms.

        """
        assert self.template._tpl is not NULL

        cdef Atom atom
        cdef int  i
        cdef int  j
        cdef int  k
        cdef int  count = self.template._tpl.count(self.template._tpl)
        cdef list atoms = []

        cdef const double* M = self._rotation
        cdef const double* c = self._centre[0]
        cdef const double* v = self._centre[1]

        for k in range(count):
            atom = Atom.__new__(Atom)
            if transform:
                atom._atom = <_Atom*> malloc(sizeof(_Atom))
                memcpy(atom._atom, &self._atoms[k], sizeof(_Atom))
                for i in range(3):
                    atom._atom.x[i] = v[i]
                    for j in range(3):
                        atom._atom.x[i] += M[3*i + j] * (self._atoms[k].x[j] - c[j])
            else:
                atom.owned = True
                atom.owner = self
                atom._atom = &self._atoms[k]

            atoms.append(atom)

        return atoms

    cpdef Molecule molecule(self, bint transform=False):
        """Get the molecule matching the template.

        Arguments:
            transform (`bool`): Whether or not to transform coordinates
                of the molecule atoms into template frame.

        Returns:
            `~pyjess.Molecule`: The matching molecule, optionally
            rotated to match the template coordinate.

        .. versionadded:: 0.5.0

        """
        assert self.template._tpl is not NULL

        cdef _Atom*        atom
        cdef Molecule      mol
        cdef size_t        i
        cdef size_t        j
        cdef size_t        k
        cdef const double* M = self._rotation
        cdef const double* c = self._centre[0]
        cdef const double* v = self._centre[1]

        if not transform:
            return self._molecule

        mol = self._molecule.copy()
        for k in range(mol._mol.count):
            atom = mol._mol.atom[k]
            for i in range(3):
                atom.x[i] = v[i]
                for j in range(3):
                    atom.x[i] += M[3*i + j] * (self._molecule._mol.atom[k].x[j] - c[j])

        return mol


cdef class Jess:
    """A handle to run Jess over a list of templates.

    .. versionadded:: 0.4.0
       Equality, hashing and pickle protocol support.

    """
    cdef _Jess* _jess
    cdef dict   _indices
    cdef tuple  _templates
    cdef size_t length

    def __cinit__(self):
        self._jess = NULL
        self.length = 0

    def __dealloc__(self):
        jess.jess.Jess_free(self._jess)

    def __init__(self, object templates = ()):
        """__init__(self, templates=())\n--\n

        Create a new Jess database containing the given templates.

        Arguments:
            templates (sequence of `~pyjess.Template`): The templates to
                index in the database for further querying.

        Caution:
            The `~pyjess.Template` objects given in argument will be copied
            because the internal C data structure requires ownership of the
            data. Modification to the original `~pyjess.Template` objects will
            not have an effect on the newly created `~pyjess.Jess` templates.

        """
        cdef Template   template
        cdef _Template* tpl
        cdef list       _templates = []

        self._jess = jess.jess.Jess_create()
        self._indices = {}

        for template in templates:
            # NOTE: the Jess storage owns the data, so we make a copy of the
            #       template given as argument to avoid a double-free.
            tpl = template._tpl.copy(template._tpl)
            jess.jess.Jess_addTemplate(self._jess, tpl)
            self._indices[<size_t> tpl] = self.length
            _templates.append(template)
            self.length += 1

        self._templates = tuple(_templates)

    def __copy__(self):
        return self.copy()

    def __reduce__(self):
        return type(self), (self._templates,)

    def __eq__(self, object other):
        cdef Jess other_
        if not isinstance(other, Jess):
            return NotImplemented
        other_ = other
        return self._templates == other_._templates

    def __hash__(self):
        return hash((Jess, self._templates))

    def __len__(self):
        return self.length

    def __getitem__(self, object index):
        cdef ssize_t index_

        if isinstance(index, slice):
            indices = range(*index.indices(self.length))
            return type(self)(map(self.__getitem__, indices))
        else:
            index_ = index
            if index_ < 0:
                index_ += self.length
            if index_ < 0 or index_ >= self.length:
                raise IndexError(index)
            return self._templates[index_]

    cpdef Jess copy(self):
        """Create a copy of the `Jess` object.

        Returns:
            `~pyjess.Jess`: A `Jess` object containing the same templates.

        .. versionadded:: 0.4.0

        """
        return type(self)(self._templates)

    def query(
        self,
        Molecule molecule,
        double rmsd_threshold,
        double distance_cutoff,
        double max_dynamic_distance,
        *,
        object max_candidates = None,
        bint ignore_chain = False,
        bint best_match = False,
        bint reorder = True,
    ):
        """Scan for templates matching the given molecule.

        Arguments:
            molecule (`~pyjess.Molecule`): The protein to match the
                templates to.
            rmsd_threshold (`float`): The RMSD threshold for reporting
                results.
            distance_cutoff (`float`): The global distance cutoff
                used to guide the search.
            max_dynamic_distance (`float`): The maximum template/query
                dynamic distance after adding the global distance cutoff
                and the individual atom distance cutoff defined for each
                atom of the template.
            max_candidates (`int` or `None`): The maximum number of candidate 
                hits to report *by template*.
            ignore_chain (`bool`): Whether to check or ignore the chain of
                the atoms to match.
            best_match (`bool`): Pass `True` to return only the best match
                to each template.
            reorder (`bool`): Whether to enable template atom reordering
                to accelerate matching in the scanner algorithm. Pass 
                `False` to reverse to the original, slower algorithm
                which matches atoms in the same order as they appear in
                the template, at the cost of longer run times.

        Returns:
            `~pyjess.Query`: An iterator over the query hits.

        Caution:
            Since ``v0.6.0``, this function uses an optimized variant of
            the Jess scanning algorithm which minimized the number of steps
            needed to generate matches, by re-ordering the order the 
            template atoms are iterated upon. Because of this change,
            the query may return *exactly* the same matches but in an order 
            that *differs* from the original Jess version. If you really
            need results in the original order, set ``reorder`` to `False`.

        .. versionadded:: 0.6.0
            The ``reorder`` argument, defaulting to `True`.

        .. versionchanged:: 0.7.0
            Default value of ``max_candidates`` argument to `None`.

        """
        cdef Query query = Query.__new__(Query)
        query.max_candidates = max_candidates
        query.ignore_chain = ignore_chain
        query.rmsd_threshold = rmsd_threshold
        query.best_match = best_match
        query.molecule = molecule
        query.jess = self
        query._jq = jess.jess.Jess_query(
            self._jess,
            molecule._mol,
            distance_cutoff,
            max_dynamic_distance,
            reorder,
        )
        return query



from .molecule import Molecule
from .template import Template


cdef extern from "Scanner.h" nogil:

    cdef struct _Scanner:
        pass
    ctypedef _Scanner Scanner

    Scanner* Scanner_create(Molecule*, Template*, double, double)
    void Scanner_free(Scanner*)
    Atom** Scanner_next(Scanner*, int)
    double Scanner_rmsd(Scanner*)




import argparse
import pathlib
import subprocess
import io
import re

RX = re.compile(r"diff --git a/([a-zA-Z\./]+) b/([a-zA-Z\./]+)")


parser = argparse.ArgumentParser()
parser.add_argument("--input", default=pathlib.Path("vendor", "jess"), type=pathlib.Path)
parser.add_argument("--base", default="testing")
parser.add_argument("--head", default="martin-update")
parser.add_argument("--output", default=pathlib.Path("patches"), type=pathlib.Path)
args = parser.parse_args()



proc = subprocess.run(["git", "checkout", args.head], cwd=args.input)
proc.check_returncode()

proc = subprocess.run(["git", "diff", args.base], cwd=args.input, capture_output=True)
proc.check_returncode()
patch = proc.stdout.decode()

proc = subprocess.run(["git", "checkout", args.base], cwd=args.input)
proc.check_returncode()



filename = None
buffer = []

for line in io.StringIO(patch):
    m = RX.match(line)
    if m:
        if filename is not None:
            with args.output.joinpath(f"{filename}.patch").open("w") as dst:
                dst.writelines(buffer)
                buffer.clear()
        path = pathlib.Path(m.group(1))
        filename = path.name
    buffer.append(line)

if filename is not None:
    with args.output.joinpath(f"{filename}.patch").open("w") as dst:
        dst.writelines(buffer)
        buffer.clear()




import argparse
import re


_HEADER_PATTERN = re.compile(r"^@@ -(\d+),?(\d+)? \+(\d+),?(\d+)? @@.*$")


def _apply_patch(s, patch, revert=False):
    # see https://stackoverflow.com/a/40967337
    s = s.splitlines(keepends=True)
    p = patch.splitlines(keepends=True)
    
    i = 0
    while not p[i].startswith(("---", "+++")):
        i += 1
    p = p[i:]
    
    t = []
    i = 0
    sl = 0
    midx, sign = (1, "+") if not revert else (3, "-")
    while i < len(p) and p[i].startswith(("---", "+++")):
        i += 1  # skip header lines

    while i < len(p):
        match = _HEADER_PATTERN.match(p[i])
        if not match:
            raise ValueError("Invalid line in patch: {!r}".format(p[i]))
        i += 1
        l = int(match.group(midx)) - 1 + (match.group(midx + 1) == "0")
        t.extend(s[sl:l])
        sl = l
        while i < len(p) and p[i][0] != "@":
            if i + 1 < len(p) and p[i + 1][0] == "\\":
                line = p[i][:-1]
                i += 2
            else:
                line = p[i]
                i += 1
            if len(line) > 0:
                if line[0] == sign or line[0] == " ":
                    t += line[1:]
                sl += line[0] != sign

    t.extend(s[sl:])
    return "".join(t)


parser = argparse.ArgumentParser()
parser.add_argument("-i", "--input", required=True)
parser.add_argument("-p", "--patch", required=True)
parser.add_argument("-o", "--output", required=True)
args = parser.parse_args()

with open(args.input, "r") as f:
    in_ = f.read()
with open(args.patch, "r") as f:
    patch = f.read()

patched = _apply_patch(in_, patch)

with open(args.output, "w") as dst:
    dst.write(patched)



import ctypes
import unittest
import pickle

from .._jess import Atom


class TestAtom(unittest.TestCase):
    
    def _create_atom(self, **kwargs):
        default = dict(serial=1, name='N', altloc=' ', residue_name='GLN', chain_id='A', residue_number=3, x=8.171, y=-51.403, z=42.886, segment='', insertion_code=' ', occupancy=1.0, temperature_factor=55.63, charge=0, element='N')
        default.update(kwargs)
        return Atom(**default)

    def test_load(self):
        atom = Atom.loads("ATOM     39  CA  PRO A 469     -14.948   2.091  10.228  1.00 27.71           C")
        self.assertEqual(atom.serial, 39)
        self.assertEqual(atom.name, 'CA')
        self.assertEqual(atom.residue_name, 'PRO')
        self.assertEqual(atom.chain_id, 'A')
        self.assertEqual(atom.residue_number, 469)
        self.assertAlmostEqual(atom.x, -14.948, places=3)
        self.assertAlmostEqual(atom.y, 2.091, places=3)
        self.assertAlmostEqual(atom.z, 10.228, places=3)
        self.assertAlmostEqual(atom.occupancy, 1.00, places=2)
        self.assertAlmostEqual(atom.temperature_factor, 27.71, places=2)
        self.assertEqual(atom.segment, '')
        self.assertEqual(atom.element, 'C')
        self.assertEqual(atom.charge, 0)

    def test_init_consistency(self):
        loaded = Atom.loads("ATOM     39  CA  PRO A 469     -14.948   2.091  10.228  1.00 27.71           C")
        created = Atom(
            serial=loaded.serial,
            name=loaded.name,
            residue_name=loaded.residue_name,
            chain_id=loaded.chain_id,
            altloc=loaded.altloc,
            insertion_code=loaded.insertion_code,
            residue_number=loaded.residue_number,
            x=loaded.x,
            y=loaded.y,
            z=loaded.z,
            occupancy=loaded.occupancy,
            temperature_factor=loaded.temperature_factor,
            element=loaded.element,
        )
        self.assertEqual(loaded, created)
        
    def test_init_invalid_chain_id(self):
        self.assertRaises(ValueError, self._create_atom, chain_id="too long")

    def test_init_invalid_name(self):
        self.assertRaises(ValueError, self._create_atom, name="too long")

    def test_init_invalid_residue_name(self):
        self.assertRaises(ValueError, self._create_atom, residue_name="too long")

    def test_hash(self):
        a1 = self._create_atom()
        a2 = self._create_atom()
        self.assertEqual(hash(a1), hash(a2))
        self.assertIsNot(a1, a2)
        a3 = self._create_atom(x=1.0)
        self.assertNotEqual(hash(a1), hash(a3))

    def test_eq(self):
        a1 = self._create_atom()
        a2 = self._create_atom()
        self.assertEqual(a1, a2)
        self.assertIsNot(a1, a2)
        a3 = self._create_atom(x=1.0)
        self.assertNotEqual(a1, a3)

    def test_repr_roundtrip(self):
        atom = self._create_atom()
        copy = eval(repr(atom))
        self.assertEqual(atom.serial, copy.serial)
        self.assertEqual(atom.altloc, copy.altloc)
        self.assertEqual(atom.name, copy.name)
        self.assertEqual(atom.residue_name, copy.residue_name)
        self.assertEqual(atom.residue_number, copy.residue_number)
        self.assertEqual(atom.element, copy.element)
        self.assertEqual(atom.insertion_code, copy.insertion_code)
        self.assertEqual(atom.chain_id, copy.chain_id)
        self.assertEqual(atom.occupancy, copy.occupancy)
        self.assertEqual(atom.temperature_factor, copy.temperature_factor)
        self.assertEqual(atom.charge, copy.charge)
        self.assertEqual(atom.x, copy.x)
        self.assertEqual(atom.y, copy.y)
        self.assertEqual(atom.z, copy.z)
        self.assertEqual(atom, copy)

    def test_pickle_roundtrip(self):
        atom = self._create_atom()
        copy = pickle.loads(pickle.dumps(atom))
        self.assertEqual(atom.serial, copy.serial)
        self.assertEqual(atom.altloc, copy.altloc)
        self.assertEqual(atom.name, copy.name)
        self.assertEqual(atom.residue_name, copy.residue_name)
        self.assertEqual(atom.residue_number, copy.residue_number)
        self.assertEqual(atom.element, copy.element)
        self.assertEqual(atom.insertion_code, copy.insertion_code)
        self.assertEqual(atom.chain_id, copy.chain_id)
        self.assertEqual(atom.occupancy, copy.occupancy)
        self.assertEqual(atom.temperature_factor, copy.temperature_factor)
        self.assertEqual(atom.charge, copy.charge)
        self.assertEqual(atom.x, copy.x)
        self.assertEqual(atom.y, copy.y)
        self.assertEqual(atom.z, copy.z)
        self.assertEqual(atom, copy)


from libc.stdio cimport FILE

from .atom cimport Atom
from .res_index cimport ResIndex


cdef extern from "Molecule.h" nogil:

    cdef struct _Molecule:
        int count
        char[5] id
        ResIndex* index
        Atom** atom

    ctypedef _Molecule Molecule

    Molecule* Molecule_create(FILE*, int, double)
    void Molecule_free(Molecule*) 
    int Molecule_count(const Molecule*)
    const Atom* Molecule_atom(const Molecule*, int)
    const char* Molecule_id(const Molecule*)


import math
import unittest
import sys
import pickle

from .._jess import Template, Molecule, Jess
from .utils import files
from . import data


class TestJess(unittest.TestCase):

    def test_init_empty(self):
        jess = Jess()
        self.assertEqual(len(jess), 0)
        self.assertFalse(bool(jess))

    def test_query_empty(self):
        mol = Molecule()
        jess = Jess()
        hits = jess.query(mol, 2, 2, 4)
        self.assertRaises(StopIteration, next, hits)

    def test_hash_empty(self):
        j1 = Jess()
        j2 = Jess()
        self.assertEqual(j1, j2)
        self.assertEqual(hash(j1), hash(j2))
        self.assertIsNot(j1, j2)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_copy(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template1 = Template.load(f)
        with files(data).joinpath("template_02.qry").open() as f:
            template2 = Template.load(f)
        jess = Jess([template1, template2])
        copy = jess.copy()
        self.assertEqual(jess, copy)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_hash(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template1 = Template.load(f)
        with files(data).joinpath("template_02.qry").open() as f:
            template2 = Template.load(f)
        j1 = Jess([template1, template2])
        j2 = Jess([template1, template2])
        self.assertEqual(hash(j1), hash(j2))
        self.assertIsNot(j1, j2)
        j3 = Jess([template1])
        self.assertNotEqual(hash(j1), hash(j3))

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_eq(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template1 = Template.load(f)
        with files(data).joinpath("template_02.qry").open() as f:
            template2 = Template.load(f)
        j1 = Jess([template1, template2])
        j2 = Jess([template1, template2])
        self.assertEqual(j1, j2)
        self.assertIsNot(j1, j2)
        j3 = Jess([template1])
        self.assertNotEqual(j1, j3)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_pickle_roundtrip(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template1 = Template.load(f)
        with files(data).joinpath("template_02.qry").open() as f:
            template2 = Template.load(f)
        jess = Jess([template1, template2])
        copy = pickle.loads(pickle.dumps(jess))
        self.assertEqual(jess, copy)

    @unittest.skipUnless(sys.implementation.name == "cpython", "only available on CPython")
    @unittest.skipUnless(files, "importlib.resources not available")
    def test_sizeof(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template = Template.load(f)
            jess = Jess([template])
        self.assertGreater(sys.getsizeof(jess), 0)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_getitem(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template = Template.load(f)
            jess = Jess([template])

        self.assertEqual(jess[0], template)
        self.assertEqual(len(jess[:1]), 1)
        self.assertEqual(len(jess[1:]), 0)


    @unittest.skipUnless(files, "importlib.resources not available")
    def test_multiple_query_split(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template1 = Template.load(f)
        with files(data).joinpath("template_02.qry").open() as f:
            template2 = Template.load(f)
        j0 = Jess([template1, template2])
        j1 = Jess([template1])
        j2 = Jess([template2])

        with files(data).joinpath("1AMY.pdb").open() as f:
            molecule = Molecule.load(f)

        h0 = list(j0.query(molecule, 2, 5, 5))
        h1 = list(j1.query(molecule, 2, 5, 5))
        h2 = list(j2.query(molecule, 2, 5, 5))

        self.assertEqual(len(h0), len(h1) + len(h2))
        for hit0, hit1 in zip(h0, (*h1, *h2)):
            self.assertEqual(hit0.atoms(), hit1.atoms())
            self.assertEqual(hit0.evalue, hit1.evalue)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_query_template_subclass(self):

        class MyTemplate(Template):
            pass
        
        with files(data).joinpath("template_01.qry").open() as f:
            template = MyTemplate.load(f)
            jess = Jess([template])
        with files(data).joinpath("pdb1lnb.pdb").open() as f:
            molecule = Molecule.load(f)

        self.assertEqual(jess[0], template)
        self.assertIsInstance(jess[0], MyTemplate)
        hits = list(jess.query(molecule, 1, 2, 2))
        self.assertIsInstance(hits[0].template, MyTemplate)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_query(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template = Template.load(f)
            jess = Jess([template])
        with files(data).joinpath("pdb1lnb.pdb").open() as f:
            molecule = Molecule.load(f)

        hits = list(jess.query(molecule, 1, 2, 2))
        self.assertEqual(len(hits), 1)
        self.assertIs(hits[0].template, template)
        self.assertAlmostEqual(hits[0].rmsd, 0.555, places=3)
        self.assertAlmostEqual(hits[0].determinant, 1.0, places=3)
        self.assertAlmostEqual(hits[0].log_evalue, -2.04, places=1)
        self.assertAlmostEqual(hits[0].evalue, math.exp(-2.04), places=1)

        hits = list(jess.query(molecule, 2, 5, 3))
        self.assertEqual(len(hits), 2)
        self.assertIs(hits[0].template, template)
        self.assertAlmostEqual(hits[0].rmsd, 0.555, places=3)
        self.assertAlmostEqual(hits[0].determinant, 1.0, places=3)
        self.assertAlmostEqual(hits[0].log_evalue, -2.04, places=1)
        self.assertAlmostEqual(hits[0].evalue, math.exp(-2.04), places=1)
        self.assertIs(hits[1].template, template)
        self.assertAlmostEqual(hits[1].rmsd, 1.440, places=3)
        self.assertAlmostEqual(hits[1].determinant, 1.0, places=3)
        self.assertAlmostEqual(hits[1].log_evalue, 0.17, places=1)
        self.assertAlmostEqual(hits[1].evalue, math.exp(0.17), places=1)

        hits = list(jess.query(molecule, 2, 5, 5))
        self.assertEqual(len(hits), 3)
        self.assertIs(hits[0].template, template)
        self.assertAlmostEqual(hits[0].rmsd, 0.555, places=3)
        self.assertAlmostEqual(hits[0].determinant, 1.0, places=3)
        self.assertAlmostEqual(hits[0].log_evalue, -2.04, places=1)
        self.assertAlmostEqual(hits[0].evalue, math.exp(-2.04), places=1)
        self.assertIs(hits[1].template, template)
        self.assertAlmostEqual(hits[1].rmsd, 1.440, places=3)
        self.assertAlmostEqual(hits[1].determinant, 1.0, places=3)
        self.assertAlmostEqual(hits[1].log_evalue, 0.17, places=1)
        self.assertAlmostEqual(hits[1].evalue, math.exp(0.17), places=1)
        self.assertIs(hits[2].template, template)
        self.assertAlmostEqual(hits[2].rmsd, 1.644, places=3)
        self.assertAlmostEqual(hits[2].determinant, 1.0, places=3)
        self.assertAlmostEqual(hits[2].log_evalue, 0.68, places=1)
        self.assertAlmostEqual(hits[2].evalue, math.exp(0.68), places=1)

        hits = list(jess.query(molecule.conserved(10.0), 1, 2, 2))
        self.assertEqual(len(hits), 1)
        self.assertIs(hits[0].template, template)
        self.assertAlmostEqual(hits[0].rmsd, 0.555, places=3)
        self.assertAlmostEqual(hits[0].determinant, 1.0, places=3)
        self.assertAlmostEqual(hits[0].log_evalue, -2.10, places=1)
        self.assertAlmostEqual(hits[0].evalue, math.exp(-2.10), places=1)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_query_best_match(self):
        with files(data).joinpath("template_01.qry").open() as f:
            template = Template.load(f)
        jess = Jess([template, template, template])
        with files(data).joinpath("pdb1lnb.pdb").open() as f:
            molecule = Molecule.load(f)

        hits = list(jess.query(molecule, 2, 5, 5, best_match=True))
        self.assertEqual(len(hits), 3)
        for hit in hits:
            self.assertIs(hit.template, template)
            self.assertAlmostEqual(hit.rmsd, 0.555, places=3)
            self.assertAlmostEqual(hit.determinant, 1.0, places=3)
            self.assertAlmostEqual(hit.log_evalue, -2.04, places=1)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_mcsa_query_no_reorder(self):
        with files(data).joinpath("1.3.3.tpl").open() as f:
            template = Template.load(f)
            jess = Jess([template])
        with files(data).joinpath("1AMY.pdb").open() as f:
            molecule = Molecule.load(f)
        with files(data).joinpath("1AMY+1.3.3.txt").open() as f:
            results = list(filter(None, f.read().split("REMARK")))

        hits = list(jess.query(molecule, 2, 4, 4, reorder=False))
        self.assertEqual(len(hits), len(results))
        for hit, block in zip(hits, results):
            self.assertIs(hit.template, template)

            lines = block.strip().splitlines()
            query_id, rmsd, template_id, _, determinant, _, logE = lines[0].split()
            self.assertEqual(query_id, "1AMY")
            self.assertAlmostEqual(float(rmsd), hit.rmsd, places=3)
            self.assertAlmostEqual(float(determinant), hit.determinant, places=1)
            self.assertAlmostEqual(float(logE), hit.log_evalue, places=1)

            atom_lines = lines[1:-1]
            atoms = hit.atoms()
            self.assertEqual(len(atoms), len(atom_lines))
            for atom, atom_line in zip(atoms, atom_lines):
                self.assertEqual(atom.serial, int(atom_line[7:12]))
                self.assertEqual(atom.name, atom_line[13:17].strip())
                self.assertEqual(atom.residue_name, atom_line[17:21].strip())
                self.assertEqual(atom.chain_id, atom_line[21:23].strip())
                self.assertEqual(atom.residue_number, int(atom_line[23:27]))
                self.assertAlmostEqual(atom.x, float(atom_line[31:39]), places=3)
                self.assertAlmostEqual(atom.y, float(atom_line[39:47]), places=3)
                self.assertAlmostEqual(atom.z, float(atom_line[47:55]), places=3)
                self.assertAlmostEqual(atom.occupancy, float(atom_line[55:61]), places=3)
                self.assertAlmostEqual(atom.temperature_factor, float(atom_line[61:67]), places=3)

            atoms = hit.atoms(transform=False)
            self.assertEqual(len(atoms), len(atom_lines))
            for atom, atom_line in zip(atoms, atom_lines):
                self.assertEqual(atom.serial, int(atom_line[7:12]))
                self.assertEqual(atom.name, atom_line[13:17].strip())
                self.assertEqual(atom.residue_name, atom_line[17:21].strip())
                self.assertEqual(atom.chain_id, atom_line[21:23].strip())
                self.assertEqual(atom.residue_number, int(atom_line[23:27]))
                self.assertAlmostEqual(atom.occupancy, float(atom_line[55:61]), places=3)
                self.assertAlmostEqual(atom.temperature_factor, float(atom_line[61:67]), places=3)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_mcsa_query_reorder(self):
        with files(data).joinpath("1.3.3.tpl").open() as f:
            template = Template.load(f)
            jess = Jess([template])
        with files(data).joinpath("1AMY.pdb").open() as f:
            molecule = Molecule.load(f)
        with files(data).joinpath("1AMY+1.3.3.txt").open() as f:
            results = list(filter(None, f.read().split("REMARK")))

        hits = list(jess.query(molecule, 2, 4, 4, reorder=True))
        self.assertEqual(len(hits), len(results))

        # `reorder=True` means that we may get results in a different order
        # to Jess, so we need to match the hits in the file by residue number
        # to make sure we compare them consistently.
        
        results_by_serials = {}
        for block in results:
            lines = block.strip().splitlines()
            serials = tuple([ int(line.split()[1]) for line in lines[1:-1] ])
            results_by_serials[serials] = block

        for hit in hits:
            self.assertIs(hit.template, template)
            block = results_by_serials[tuple(atom.serial for atom in hit.atoms(False))]

            lines = block.strip().splitlines()
            query_id, rmsd, template_id, _, determinant, _, logE = lines[0].split()
            self.assertEqual(query_id, "1AMY")
            self.assertAlmostEqual(float(rmsd), hit.rmsd, places=3)
            self.assertAlmostEqual(float(determinant), hit.determinant, places=1)
            self.assertAlmostEqual(float(logE), hit.log_evalue, places=1)

            atom_lines = lines[1:-1]
            atoms = hit.atoms()
            self.assertEqual(len(atoms), len(atom_lines))
            for atom, atom_line in zip(atoms, atom_lines):
                self.assertEqual(atom.serial, int(atom_line[7:12]))
                self.assertEqual(atom.name, atom_line[13:17].strip())
                self.assertEqual(atom.residue_name, atom_line[17:21].strip())
                self.assertEqual(atom.chain_id, atom_line[21:23].strip())
                self.assertEqual(atom.residue_number, int(atom_line[23:27]))
                self.assertAlmostEqual(atom.x, float(atom_line[31:39]), places=3)
                self.assertAlmostEqual(atom.y, float(atom_line[39:47]), places=3)
                self.assertAlmostEqual(atom.z, float(atom_line[47:55]), places=3)
                self.assertAlmostEqual(atom.occupancy, float(atom_line[55:61]), places=3)
                self.assertAlmostEqual(atom.temperature_factor, float(atom_line[61:67]), places=3)

            atoms = hit.atoms(transform=False)
            self.assertEqual(len(atoms), len(atom_lines))
            for atom, atom_line in zip(atoms, atom_lines):
                self.assertEqual(atom.serial, int(atom_line[7:12]))
                self.assertEqual(atom.name, atom_line[13:17].strip())
                self.assertEqual(atom.residue_name, atom_line[17:21].strip())
                self.assertEqual(atom.chain_id, atom_line[21:23].strip())
                self.assertEqual(atom.residue_number, int(atom_line[23:27]))
                self.assertAlmostEqual(atom.occupancy, float(atom_line[55:61]), places=3)
                self.assertAlmostEqual(atom.temperature_factor, float(atom_line[61:67]), places=3)




cdef extern from "Region.h" nogil:

    struct _Region:
        int (*intersectionQ)(Region*, double*, double*, int)
        int (*inclusionQ)(Region*, double*, int)
        void (*free)(Region*)
    ctypedef _Region Region

    double Region_volume(Region*, double, double*, double*, int)


from .atom cimport Atom
from .molecule cimport Molecule


cdef extern from "CandidateSet.h" nogil:

    struct _CandidateSet:
        int count
        Atom **atom
        double **coord

    ctypedef _CandidateSet CandidateSet

    CandidateSet *CandidateSet_create(const Molecule *M)
    void CandidateSet_free(CandidateSet*)


from .atom cimport Atom

cdef extern from "TessAtom.h" nogil:

    struct _TessAtom:
        int code
        int resSeq
        int nameCount
        int resNameCount
        char chainID1
        char chainID2
        char **name
        char **resName
        double pos[3]
        double distWeight

    ctypedef _TessAtom TessAtom

    TessAtom* TessAtom_create(const char*)
    void TessAtom_free(TessAtom*)
    const double* TessAtom_position(const TessAtom*)
    int TessAtom_match(const TessAtom*, const Atom*)
    int TessAtom_resSeq(const TessAtom*)
    char TessAtom_chainID1(const TessAtom*)
    char TessAtom_chainID2(const TessAtom*)
    double TessAtom_distWeight(const TessAtom*)
    
    TessAtom* TessAtom_copy(const TessAtom*)
    double TessAtom_distance(const TessAtom*, const TessAtom*) 


from .region cimport Region


cdef extern from "KdTree.h" nogil:

    cdef struct _KdTree:
        pass
    cdef struct _KdTreeQuery:
        pass

    ctypedef _KdTree KdTree
    ctypedef _KdTreeQuery KdTreeQuery

    KdTree* KdTree_create(double**, int, int)
    void KdTree_free(KdTree**)
    KdTreeQuery* KdTree_query(KdTree*, Region*)

    void KdTreeQuery_free(KdTreeQuery*)
    int KdTreeQuery_next(KdTreeQuery*)


from .atom cimport Atom
from .molecule cimport Molecule
from .super cimport Superposition
from .template cimport Template


cdef extern from "Jess.h" nogil:

    ctypedef struct _Jess:
        pass

    ctypedef struct _JessQuery:
        pass

    ctypedef _Jess Jess
    ctypedef _JessQuery JessQuery

    Jess* Jess_create()
    void Jess_free(Jess*)
    void Jess_addTemplate(Jess*, Template*)
    JessQuery* Jess_query(Jess*, Molecule*, double, double, bint)

    void JessQuery_free(JessQuery*)
    int JessQuery_next(JessQuery*, int)
    int JessQuery_nextTemplate(JessQuery*)
    Template* JessQuery_template(JessQuery*)
    const Molecule* JessQuery_molecule(JessQuery*)
    Atom** JessQuery_atoms(JessQuery*)
    Superposition* JessQuery_superposition(JessQuery*)





from .region cimport Region


cdef extern from "Join.h" nogil:

    ctypedef enum JoinType:
        innerJoin
        outerJoin

    Region* Join_create(Region**, int, JoinType)
    void Join_free(Region*)




from .atom cimport Atom


cdef extern from "ResIndex.h" nogil:

    struct _ResIndex:
        int count
        Atom **atom
        double **coord

    ctypedef _ResIndex ResIndex

    extern ResIndex* ResIndex_create(Atom**, int)
    extern Atom** ResIndex_get(ResIndex*, const char[4])
    extern void ResIndex_free(ResIndex*)


import os
import typing
from typing import Any, Generic, Union, Optional, Sequence, Iterator, Iterable, List, TextIO, Sized, TypeVar, Tuple

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

MATCH_MODE = Literal[
    -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 100, 101, 102, 103, 104, 105, 106, 107
]

_SELF = TypeVar("_SELF")

__version__: str

class Molecule(Sequence[Atom]):
    @classmethod
    def load(cls, file: Union[TextIO, str, os.PathLike[str]], id: Optional[str] = None, ignore_endmdl: bool = False) -> Molecule: ...
    @classmethod
    def loads(cls, text: str, id: Optional[str] = None, ignore_endmdl: bool = False) -> Molecule: ...
    def __init__(self, atoms: Sequence[Atom] = (), id: Optional[str] = None): ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> Atom: ...
    @typing.overload
    def __getitem__(self: _SELF, index: slice) -> _SELF: ...
    @typing.overload
    def __getitem__(self: _SELF, index: Union[int, slice]) -> Union[Atom, _SELF]: ...
    def __copy__(self) -> Molecule: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __reduce__(self) -> Tuple[Any, ...]: ...
    @property
    def id(self) -> Optional[str]: ...
    def conserved(self: _SELF, cutoff: float = 0.0) -> _SELF: ...
    def copy(self) -> Molecule: ...

class Atom:
    @classmethod
    def load(cls, file: Union[TextIO, str, os.PathLike[str]]) -> Atom: ...
    @classmethod
    def loads(cls, text: str) -> Atom: ...
    def __init__(
        self,
        *,
        serial: int,
        name: str,
        altloc: str,
        residue_name: str,
        chain_id: str,
        residue_number: int,
        insertion_code: str,
        x: float,
        y: float,
        z: float,
        occupancy: float = 0.0,
        temperature_factor: float = 0.0,
        segment: str = "",
        element: str = "",
        charge: int = 0,
    ): ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __reduce__(self) -> Tuple[Any, ...]: ...
    def __repr__(self) -> str: ...
    def __copy__(self) -> Atom: ...
    @property
    def serial(self) -> int: ...
    @property
    def altloc(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def residue_name(self) -> str: ...
    @property
    def residue_number(self) -> int: ...
    @property
    def segment(self) -> str: ...
    @property
    def element(self) -> str: ...
    @property
    def insertion_code(self) -> str: ...
    @property
    def chain_id(self) -> str: ...
    @property
    def occupancy(self) -> float: ...
    @property
    def temperature_factor(self) -> float: ...
    @property
    def charge(self) -> int: ...
    @property
    def x(self) -> float: ...
    @property
    def y(self) -> float: ...
    @property
    def z(self) -> float: ...
    def copy(self) -> Atom: ...

class TemplateAtom:
    @classmethod
    def load(cls, file: Union[TextIO, str, os.PathLike[str]]) -> Atom: ...
    @classmethod
    def loads(cls, text: str) -> TemplateAtom: ...
    def __init__(
        self,
        *,
        chain_id: str,
        residue_number: int,
        x: float,
        y: float,
        z: float,
        residue_names: Sequence[str],
        atom_names: Sequence[str],
        distance_weight: float = 0.0,
        match_mode: MATCH_MODE = 0,
    ): ...
    def __repr__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __reduce__(self) -> Tuple[Any, ...]: ...
    def __copy__(self: _SELF) -> _SELF: ...
    @property
    def match_mode(self) -> MATCH_MODE: ...
    @property
    def residue_number(self) -> int: ...
    @property
    def chain_id(self) -> str: ...
    @property
    def x(self) -> float: ...
    @property
    def y(self) -> float: ...
    @property
    def z(self) -> float: ...
    @property
    def atom_names(self) -> Tuple[str, ...]: ...
    @property
    def residue_names(self) -> Tuple[str, ...]: ...
    @property
    def distance_weight(self) -> float: ...
    def copy(self: _SELF) -> _SELF: ...

class Template(Sequence[TemplateAtom]):
    @classmethod
    def load(cls, file: Union[TextIO, str, os.PathLike[str]], id: Optional[str] = None) -> Template: ...
    @classmethod
    def loads(cls, text: str, id: Optional[str] = None) -> Template: ...
    def __init__(self, atoms: Sequence[TemplateAtom] = (), id: Optional[str] = None): ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> TemplateAtom: ...
    @typing.overload
    def __getitem__(self: _SELF, index: slice) -> _SELF: ...
    @typing.overload
    def __getitem__(self: _SELF, index: Union[int, slice]) -> Union[TemplateAtom, _SELF]: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __reduce__(self) -> Tuple[Any, ...]: ...
    def __copy__(self) -> Template: ...
    @property
    def id(self) -> str: ...
    @property
    def dimension(self) -> int: ...
    def copy(self) -> Template: ...

_T = TypeVar("_T", bound=Template)

class Query(Generic[_T], Iterator[Hit[_T]]):
    @property
    def jess(self) -> Jess[_T]: ...
    @property
    def molecule(self) -> Molecule: ...
    @property
    def ignore_chain(self) -> bool: ...
    @property
    def rmsd_threshold(self) -> float: ...
    @property
    def max_candidates(self) -> Optional[int]: ...
    @property
    def best_match(self) -> bool: ...
    def __iter__(self) -> Query[_T]: ...
    def __next__(self) -> Hit[_T]: ...

class Hit(Generic[_T]):
    def __getstate__(self) -> Dict[str, object]: ...
    def __setstate__(self, state: Dict[str, object]) -> None: ...
    @property
    def rmsd(self) -> float: ...
    @property
    def template(self) -> _T: ...
    @property
    def determinant(self) -> float: ...
    @property
    def log_evalue(self) -> float: ...
    @property
    def evalue(self) -> float: ...
    def atoms(self, transform: bool = True) -> List[Atom]: ...
    def molecule(self, transform: bool = False) -> Molecule: ...

class Jess(Generic[_T], Sequence[_T]):
    def __init__(self, templates: Iterable[_T] = ()): ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> _T: ...
    @typing.overload
    def __getitem__(self, index: slice) -> Jess[_T]: ...
    @typing.overload
    def __getitem__(self, index: Union[int, slice]) -> Union[_T, Jess[_T]]: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __copy__(self: _SELF) -> _SELF: ...
    def __reduce__(self) -> Tuple[Any, ...]: ...
    def query(
        self,
        molecule: Molecule,
        rmsd_threshold: float,
        distance_cutoff: float,
        max_dynamic_distance: float,
        *,
        max_candidates: Optional[int] = None,
        ignore_chain: bool = False,
        best_match: bool = False,
        reorder: bool = True,
    ) -> Query[_T]: ...



cdef extern from "Atom.h" nogil:

    struct _Atom:
        int serial
        char[5] name
        char altLoc
        char[4] resName
        char chainID1
        char chainID2
        int resSeq
        char iCode
        double[3] x
        double occupancy
        double tempFactor
        char[4] segID
        char[3] element
        int charge

    ctypedef _Atom Atom

    int Atom_parse(Atom*, const char*)


from . import (
    test_atom,
    test_hit,
    test_jess,
    test_molecule,
    test_template_atom,
    test_template
)

def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_atom))
    suite.addTests(loader.loadTestsFromModule(test_hit))
    suite.addTests(loader.loadTestsFromModule(test_jess))
    suite.addTests(loader.loadTestsFromModule(test_molecule))
    suite.addTests(loader.loadTestsFromModule(test_template_atom))
    suite.addTests(loader.loadTestsFromModule(test_template))
    # test_doctest.load_tests(loader, suite, pattern)
    return suite



import unittest
import sys
import pickle

from .._jess import TemplateAtom


class TestTemplateAtom(unittest.TestCase):

    def test_load(self):
        atom = TemplateAtom.loads("ATOM      1  NE  ARG A1136       3.953   0.597  -1.721 K")
        self.assertEqual(atom.match_mode, 1)
        self.assertEqual(atom.atom_names, ("NE",))
        self.assertEqual(atom.residue_names, ("ARG", "LYS",))
        self.assertEqual(atom.chain_id, "A")
        self.assertEqual(atom.residue_number, 1136)
        self.assertEqual(atom.x, 3.953)
        self.assertEqual(atom.y, 0.597)
        self.assertEqual(atom.z, -1.721)

    def _create_atom(self, **kwargs):
        default = {
            "atom_names": ["NE"],
            "residue_names": ["ARG"],
            "chain_id": "A",
            "residue_number": 1,
            "x": 0.0,
            "y": 0.0,
            "z": 0.0,
            "match_mode": 0,
        }
        default.update(kwargs)
        return TemplateAtom(**default)

    def test_hash(self):
        a1 = self._create_atom()
        a2 = self._create_atom()
        self.assertEqual(hash(a1), hash(a2))
        self.assertIsNot(a1, a2)
        a3 = self._create_atom(x=1.0)
        self.assertNotEqual(hash(a1), hash(a3))

    def test_eq(self):
        a1 = self._create_atom()
        a2 = self._create_atom()
        self.assertEqual(a1, a2)
        self.assertIsNot(a1, a2)
        a3 = self._create_atom(x=1.0)
        self.assertNotEqual(a1, a3)

    @unittest.skipUnless(sys.implementation.name == "cpython", "only available on CPython")
    def test_sizeof(self):
        atom = self._create_atom()
        self.assertGreater(sys.getsizeof(atom), 0)

    def test_init_invalid_match_code(self):
        self.assertRaises(ValueError, self._create_atom, match_mode=2000)
        self.assertRaises(ValueError, self._create_atom, match_mode=-10)

    def test_init_invalid_residue_name(self):
        self.assertRaises(ValueError, self._create_atom, residue_names=["something"])

    def test_init_consistency(self):
        loaded = TemplateAtom.loads("ATOM      1  NE  ARG A1136       3.953   0.597  -1.721 K")
        created = TemplateAtom(
            chain_id=loaded.chain_id, 
            residue_number=loaded.residue_number,
            x=loaded.x,
            y=loaded.y,
            z=loaded.z,
            residue_names=loaded.residue_names,
            atom_names=loaded.atom_names,
            distance_weight=loaded.distance_weight,
            match_mode=loaded.match_mode
        )
        for attribute in ("atom_names", "residue_names", "chain_id", "x", "y", "z", "match_mode"):
            self.assertEqual(getattr(loaded, attribute), getattr(created, attribute))
        self.assertEqual(loaded, created)

    def test_repr_roundtrip(self):
        atom = self._create_atom()
        copy = eval(repr(atom))
        for attribute in ("atom_names", "residue_names", "chain_id", "x", "y", "z", "match_mode"):
            self.assertEqual(getattr(copy, attribute), getattr(atom, attribute))
        self.assertEqual(atom, copy)

    def test_pickle_roundtrip(self):
        atom = self._create_atom()
        copy = pickle.loads(pickle.dumps(atom))
        for attribute in ("atom_names", "residue_names", "chain_id", "x", "y", "z", "match_mode"):
            self.assertEqual(getattr(copy, attribute), getattr(atom, attribute))
        self.assertEqual(atom, copy)


import os
import pickle
import unittest
import tempfile
import textwrap
import sys

from .._jess import Atom, Molecule

MOLECULE = textwrap.dedent(
    """
    HEADER    DNA RECOMBINATION                       05-DEC-97   1A0P              
    COMPND    MOL_ID: 1;                                                            
    ATOM      1  N   GLN A   3       8.171 -51.403  42.886  1.00 55.63           N
    ATOM      2  CA  GLN A   3       9.475 -50.697  42.743  1.00 56.29           C
    ATOM      3  C   GLN A   3      10.215 -51.213  41.516  1.00 55.54           C
    ATOM      4  O   GLN A   3      10.401 -50.398  40.585  1.00 56.57           O
    ATOM      5  CB  GLN A   3      10.267 -50.747  44.040  1.00 72.29           C
    """
).strip()


class TestMolecule(unittest.TestCase):

    def _create_atom(self, **kwargs):
        default = dict(serial=1, name='N', altloc=' ', residue_name='GLN', chain_id='A', residue_number=3, x=8.171, y=-51.403, z=42.886, segment='', insertion_code=' ', occupancy=1.0, temperature_factor=55.63, charge=0, element='N')
        default.update(kwargs)
        return Atom(**default)

    def test_loads(self):
        molecule = Molecule.loads(MOLECULE)
        self.assertEqual(len(molecule), 5)
        self.assertEqual(molecule.id, '1A0P')

    @unittest.skipUnless(sys.implementation.name == "cpython", "only available on CPython")
    def test_sizeof(self):
        molecule = Molecule.loads(MOLECULE)
        self.assertGreater(sys.getsizeof(molecule), 0)

    @unittest.skipIf(os.name == "nt", "permission errors on Windows")
    def test_load_filename(self):
        with tempfile.NamedTemporaryFile("w", suffix=".pdb") as f:
            f.write(MOLECULE)
            f.flush()
            molecule = Molecule.load(f.name)
        self.assertEqual(len(molecule), 5)
        self.assertEqual(molecule.id, '1A0P')

    @unittest.skipIf(os.name == "nt", "permission errors on Windows")
    def test_load_file(self):
        with tempfile.NamedTemporaryFile("r+", suffix=".pdb") as f:
            f.write(MOLECULE)
            f.flush()
            f.seek(0)
            molecule = Molecule.load(f)
        self.assertEqual(len(molecule), 5)
        self.assertEqual(molecule.id, '1A0P')

    @unittest.skipIf(os.name == "nt", "permission errors on Windows")
    def test_load_error(self):
        self.assertRaises(FileNotFoundError, Molecule.load, "/some/nonsensical/file")
        self.assertRaises(IsADirectoryError, Molecule.load, os.path.dirname(__file__))

    def test_init(self):
        atoms = [
            self._create_atom(serial=1, name='N'),
            self._create_atom(serial=2, name='CA'),
            self._create_atom(serial=3, name='C'),
            self._create_atom(serial=4, name='O'),
        ]
        molecule = Molecule(atoms)
        self.assertEqual(molecule[0].name, 'N')
        self.assertEqual(molecule[1].name, 'CA')
        self.assertEqual(molecule[2].name, 'C')
        self.assertEqual(molecule[3].name, 'O')

    def test_init_long_id(self):
        mol = Molecule.loads(MOLECULE, id="long identifier")
        self.assertEqual(mol.id, "long identifier")

    def test_getitem_slicing(self):
        mol = Molecule.loads(MOLECULE)
        mol2 = mol[1:3]
        self.assertEqual(mol2.id, mol.id)
        self.assertEqual(len(mol2), 2)
        self.assertEqual(mol2[0].name, "CA")
        self.assertEqual(mol2[1].name, "C")

    def test_hash(self):
        atoms = [
            self._create_atom(serial=1, name='N'),
            self._create_atom(serial=2, name='CA'),
            self._create_atom(serial=3, name='C'),
            self._create_atom(serial=4, name='O'),
        ]
        mol1 = Molecule(atoms)
        mol2 = Molecule(atoms)
        self.assertEqual(hash(mol1), hash(mol2))
        self.assertIsNot(mol1, mol2)
        mol3 = Molecule(atoms[:-1])
        self.assertNotEqual(hash(mol1), hash(mol3))

    def test_eq(self):
        atoms = [
            self._create_atom(serial=1, name='N'),
            self._create_atom(serial=2, name='CA'),
            self._create_atom(serial=3, name='C'),
            self._create_atom(serial=4, name='O'),
        ]
        mol1 = Molecule(atoms)
        mol2 = Molecule(atoms)
        self.assertEqual(mol1, mol2)
        self.assertIsNot(mol1, mol2)
        mol3 = Molecule(atoms[:-1])
        self.assertNotEqual(mol1, mol3)

    def test_copy(self):
        atoms = [
            self._create_atom(serial=1, name='N'),
            self._create_atom(serial=2, name='CA'),
            self._create_atom(serial=3, name='C'),
            self._create_atom(serial=4, name='O'),
        ]
        mol1 = Molecule(atoms)
        mol2 = mol1.copy()
        self.assertEqual(list(mol1), list(mol2))
        self.assertEqual(mol1.id, mol2.id)
        self.assertEqual(mol1, mol2)

    def test_pickle_roundtrip(self):
        atoms = [
            self._create_atom(serial=1, name='N'),
            self._create_atom(serial=2, name='CA'),
            self._create_atom(serial=3, name='C'),
            self._create_atom(serial=4, name='O'),
        ]
        mol1 = Molecule(atoms)
        mol2 = pickle.loads(pickle.dumps(mol1))
        self.assertEqual(list(mol1), list(mol2))
        self.assertEqual(mol1.id, mol2.id)
        self.assertEqual(mol1, mol2)



# noqa: D104
from . import _jess
from ._jess import Atom, Hit, Jess, Query, Molecule, Template, TemplateAtom

__author__ = "Martin Larralde <martin.larralde@embl.de>"
__all__ = ["Atom", "Hit", "Jess", "Query", "Molecule", "Template", "TemplateAtom"]
__doc__ = _jess.__doc__
__version__ = _jess.__version__

# Small addition to the docstring: we want to show a link redirecting to the
# rendered version of the documentation, but this can only work when Python
# is running with docstrings enabled
if __doc__ is not None:
    __doc__ += """See Also:
    An online rendered version of the documentation for this version
    of the library on
    `Read The Docs <https://pyjess.readthedocs.io/en/v{}/>`_.

    """.format(
        __version__
    )



cdef extern from "Super.h" nogil:

    cdef struct _Superposition:
        pass
    ctypedef _Superposition Superposition

    Superposition* Superposition_create()
    void Superposition_free(Superposition*)
    void Superposition_align(Superposition*, const double*, const double*)
    int Superposition_count(const Superposition*)
    double Superposition_rmsd(Superposition*)
    double Superposition_rmsd100(Superposition*)
    const double* Superposition_centroid(Superposition*, int)
    const double* Superposition_rotation(Superposition*)


#!/usr/bin/env python3

import sys
import re
from collections import defaultdict

# Catalytic residue functional atoms definitions
RESIDUE_DEFINITIONS = { 
    'Main': {'CA': [100, ''], 'C': [100, ''], 'O': [100, '']},
    'Ptm': {'CA': [100, ''], 'C': [100, ''],  'CB': [100, '']},
    'Ala': {'C' : [0, ''], 'CA' : [0, ''], 'CB' : [0, '']},
    'Cys': {'CA' : [0, ''], 'CB' : [0, ''], 'SG' : [0, '']},
    'Asp': {'CG': [3, 'E'], 'OD1': [3, 'E'], 'OD2': [3, 'E']},
    'Glu': {'CD': [3, 'D'], 'OE1': [3, 'D'], 'OE2': [3, 'D']},
    'Phe': {'CE1': [3, ''], 'CE2': [3, ''], 'CZ': [0, '']},
    'Gly': {'CA' : [0, ''], 'C' : [0, ''], 'O' : [0, '']},
    'His': {'CG': [0, ''], 'CD2':[8, ''], 'ND1': [8, '']},
    'Ile': {'CA': [0, 'VL'], 'CB': [0, 'VL'], 'CG1': [3, 'VL']},
    'Lys': {'CD' : [0, ''], 'CE' : [0, ''], 'NZ' : [0, '']},
    'Leu': {'CA': [0, 'VI'], 'CB': [0, 'VI'], 'CG': [3, 'VI']},
    'Met': {'CG' : [0, ''], 'SD' : [0, ''], 'CE' : [0, '']},
    'Asn': {'CG': [3, 'Q'], 'OD1': [1, 'Q'] , 'ND2': [1, 'Q']},
    'Pro': {'C' : [0, ''], 'CA' : [0, ''], 'O' : [0, '']},
    'Gln': {'CD': [3, 'N'], 'OE1': [1, 'N'], 'NE2': [1, 'N']},
    'Arg': {'CZ': [0, ''], 'NH1': [3, ''], 'NH2': [3, '']},
    'Ser': {'CA': [0, 'TY'], 'CB':[0, 'TY'], 'OG': [1, 'TY']},
    'Thr': {'CA': [0, 'SY'], 'CB': [0, 'SY'], 'OG1': [1, 'SY']},
    'Trp': {'NE1' : [0, ''], 'CZ2' : [0, ''], 'CH2' : [0, '']},
    'Tyr': {'CE1': [3, 'ST'], 'CZ': [3, 'ST'], 'OH': [1, 'ST']},
    'Val': {'CA': [0, 'LI'], 'CB': [0, 'LI'], 'CG': [3, 'LI']},
    }


class Match():
    """Jess match"""

    def __init__(self, target='', template='', rmsd=None, score=None, remark_string='REMARK MATCH'):
        self.target = target
        self.template = template
        self.rmsd = float(rmsd)
        self.score = float(score)
        self.remark_string = remark_string
        self.atoms = []
        self.residues = []
        self.chains = set()

    def __repr__(self):
        return '{}\n{}\nENDMDL\n'.format(self.remark_string, '\n'.join(self.atoms))

    @property
    def id(self):
        return '{};{};{}'.format(target, template, ':'.join(list(chains)))

    @classmethod
    def from_remark(cls, remark_string):
        fields = remark_string.strip().split()
        return cls(fields[1], fields[3], fields[2], fields[-1], remark_string=remark_string)

    def add_atom(self, line):
        self.atoms.append(line)
        self.chains.add(line[20:22].strip())

    def add_residue(self, line):
        resname = line[17:20].strip()
        chain = line[20:22].strip()
        resid = line[22:26].strip()
        res = f'{resname}_{chain}_{resid}'
        if res not in self.residues:
            self.residues.append(res)

    def is_from_same_structure(self, other):
        return (self.target == other.target and self.template == other.template)

    def is_from_same_residues(self, other):
        return self.residues == other.residues

    def is_from_same_chains(self, other):
        return self.chains == other.chains

def get_highest_scored_match(matches):
    min_score = 999
    best_match = None
    best_matches = []
    for match in matches:
        if match.score < min_score:
            min_score = match.score
            best_match = match
    return best_match

def print_best_match(matches, all_instances=False):
    unique_matches = []
    for match_set in matches.values():
        best_match = get_highest_scored_match(match_set)
        # This is to discard self matches when cross-comparing templates
        if best_match.template == best_match.target:
            continue
        unique_matches.append(best_match)
    if all_instances:
        for unique_match in unique_matches:
            print(unique_match)
    else:
        best_match = get_highest_scored_match(unique_matches)
        if best_match is not None:
            print(best_match)
    return

def main(all_instances=True):
    """all_distances: If True, then all unique matches from diffent chains 
    are returned"""
    prev_match = None
    matches = defaultdict(list)
    unique_matches = []
    for line in sys.stdin:
        line=line.strip()
        if line.startswith('REMARK'):
            match = Match.from_remark(line)
        if line.startswith('ATOM') or line.startswith('HETATM'):
            match.add_atom(line)
            match.add_residue(line)
        if line.startswith('ENDMDL'):
            if prev_match and not match.is_from_same_structure(prev_match):
                print_best_match(matches, all_instances)
                matches = defaultdict(list)
            matches[':'.join(list(match.residues))].append(match)
            prev_match = match
    print_best_match(matches, all_instances)
    return

if __name__ == '__main__':
    main()



import os
import pickle
import unittest
import tempfile
import textwrap
import sys

from .._jess import Template
from .utils import files
from . import data

TEMPLATE = textwrap.dedent(
    """
    ATOM      0  NZ  LYS A1030       1.431  -9.717  -1.722
    ATOM      0  CG  ASP A1132       0.000   0.000   0.000
    ATOM      0  OD1 ASP A1132       0.595   1.085   0.000
    ATOM      0  OD2 ASP A1132       0.609  -1.111   0.000
    ATOM      1  NE  ARG A1136       3.953   0.597  -1.721 K
    ATOM      0  CG  ASN A1137       0.736  -1.869  -4.261
    ATOM      0  OD1 ASN A1137       1.760  -2.435  -4.344
    ATOM      0  ND2 ASN A1137       0.264  -1.477  -3.134
    ATOM      0  CG  ASP A1150      -0.010  -5.408  -2.070
    ATOM      0  OD1 ASP A1150      -0.587  -5.769  -1.153
    ATOM      0  OD2 ASP A1150       1.085  -5.212  -2.011
    """
).strip()


class TestTemplate(unittest.TestCase):

    def test_loads(self):
        template = Template.loads(TEMPLATE)
        self.assertEqual(len(template), 11)
        self.assertEqual(template.dimension, 5)
        self.assertEqual(template[0].residue_names, ("LYS",))
        self.assertEqual(template[0].atom_names, ("NZ",))
        self.assertEqual(template[1].atom_names, ("CG",))
        self.assertEqual(template[2].residue_number, 1132)
        self.assertEqual(template[-1].residue_number, 1150)

    @unittest.skipIf(os.name == "nt", "permission errors on Windows")
    def test_load_filename(self):
        with tempfile.NamedTemporaryFile("w", suffix=".pdb") as f:
            f.write(TEMPLATE)
            f.flush()
            template = Template.load(f.name)
        self.assertEqual(template[0].residue_names, ("LYS",))
        self.assertEqual(template[0].atom_names, ("NZ",))
        self.assertEqual(template[1].atom_names, ("CG",))
        self.assertEqual(template[2].residue_number, 1132)
        self.assertEqual(template[-1].residue_number, 1150)

    @unittest.skipIf(os.name == "nt", "permission errors on Windows")
    def test_load_file(self):
        with tempfile.NamedTemporaryFile("r+", suffix=".pdb") as f:
            f.write(TEMPLATE)
            f.flush()
            f.seek(0)
            template = Template.load(f)
        self.assertEqual(template[0].residue_names, ("LYS",))
        self.assertEqual(template[0].atom_names, ("NZ",))
        self.assertEqual(template[1].atom_names, ("CG",))
        self.assertEqual(template[2].residue_number, 1132)
        self.assertEqual(template[-1].residue_number, 1150)

    @unittest.skipIf(os.name == "nt", "permission errors on Windows")
    def test_load_error(self):
        self.assertRaises(FileNotFoundError, Template.load, "/some/nonsensical/file")
        self.assertRaises(IsADirectoryError, Template.load, os.path.dirname(__file__))

    @unittest.skipUnless(sys.implementation.name == "cpython", "only available on CPython")
    def test_sizeof(self):
        template = Template.loads(TEMPLATE)
        self.assertGreater(sys.getsizeof(template), 0)

    def test_init_empty(self):
        template = Template()
        self.assertEqual(len(template), 0)
        self.assertFalse(bool(template))

    def test_loads_id(self):
        template = Template.loads(TEMPLATE)
        self.assertIs(template.id, None)
        template = Template.loads("REMARK PDB_ID 2bw4\n" + TEMPLATE)
        self.assertEqual(template.id, "2bw4")

    def test_getitem_slicing(self):
        tpl1 = Template.loads(TEMPLATE, id="tpl1")
        tpl2 = tpl1[1:4]
        self.assertEqual(len(tpl2), 3)

    def test_hash(self):
        tpl1 = Template.loads(TEMPLATE, id="tpl1")
        tpl2 = Template.loads(TEMPLATE, id="tpl1")
        self.assertEqual(tpl1, tpl2)
        self.assertEqual(hash(tpl1), hash(tpl2))
        self.assertIsNot(tpl1, tpl2)
        tpl3 = Template.loads(TEMPLATE, id="tpl3")
        self.assertNotEqual(hash(tpl1), hash(tpl3))

    def test_eq(self):
        tpl1 = Template.loads(TEMPLATE, id="tpl1")
        tpl2 = Template.loads(TEMPLATE, id="tpl1")
        self.assertEqual(tpl1, tpl2)
        self.assertIsNot(tpl1, tpl2)
        tpl3 = Template.loads(TEMPLATE, id="tpl3")
        self.assertNotEqual(tpl1, tpl3)

    def test_copy(self):
        tpl1 = Template.loads(TEMPLATE, id="tpl1")
        tpl2 = tpl1.copy()
        self.assertEqual(len(tpl1), len(tpl2))
        self.assertEqual(tpl1, tpl2)

    def test_pickle_roundtrip(self):
        tpl1 = Template.loads(TEMPLATE, id="tpl1")
        tpl2 = pickle.loads(pickle.dumps(tpl1))
        self.assertEqual(len(tpl1), len(tpl2))
        self.assertEqual(tpl1, tpl2)

    @unittest.skipUnless(files, "importlib.resources not available")
    def test_dimension_multiple_chains(self):
        with files(data).joinpath("1sur.qry").open() as f:
            template1 = Template.load(f)
        self.assertEqual(template1.dimension, 4)
